/*
 *
 *  Child List Implementation
 *  by Tom Epperly
 *  Version: $Revision: 1.18 $
 *  Version control file: $RCSfile: child.c,v $
 *  Date last modified: $Date: 97/09/04 01:09:46 $
 *  Last modified by: $Author: ballan $
 *
 *  This file is part of the Ascend Language Interpreter.
 *
 *  Copyright (C) 1990, 1993, 1994 Thomas Guthrie Epperly
 *
 *  The Ascend Language Interpreter is free software; you can redistribute
 *  it and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  The Ascend Language Interpreter is distributed in hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge, MA 02139 USA.  Check the file named
 *  COPYING.
 *
 *  Implementation of Child list stuff.
 */

#include "utilities/ascConfig.h"
#include "utilities/ascMalloc.h"
#include "general/list.h"
#include "compiler/compiler.h"
#include "compiler/instance_enum.h"
#include "compiler/fractions.h"
#include "compiler/dimen.h"
#include "compiler/types.h"
#include "compiler/stattypes.h"
#include "compiler/statio.h"
#include "compiler/child.h"
#include "compiler/type_desc.h"
#include "compiler/cmpfunc.h"

#ifndef lint
static CONST char ChildListID[] = "$Id: child.c,v 1.18 97/09/04 01:09:46 ballan Exp $";
#endif

/* origin of sanity (or sanity of origin) check. dependent on header */
#define LegalOrigin(n) ( ((n)>0 && (n) <=8) ? (n) : 0 )

/* return nth element of a child list l */
#define CGET(l,n) \
 ((struct ChildListEntry *)gl_fetch(((CONST struct gl_list_t *)l),(n)))

/* the following will always amount to an append */
#define CPUT(l,p) gl_insert_sorted((l),(p),(CmpFunc)CmpChildListEntries)

/* copy the contents of a entry pointer to another ptr */
#define CLCOPY(src) copylistentry(src)

/* replace the following with a pool */
#define CLPOOLALLOC ((struct ChildListEntry *) \
  ascmalloc(sizeof(struct ChildListEntry)))
#define CLPOOLFREE(p) ascfree(p)

/* ok, for now it's convenient to have gllist and childlist equivalent.
 * Playing more advanced compiler games this is likely to no longer be
 * the case, so it's tightly wrapped in the ChildListPtr.
 * This also allows us tight control on the ChildListEntry memory,
 * which we need.
 */
struct ChildListStructure {
  struct gl_list_t list;
};

/* macro to fetch the gl list from the ChildListPtr */
#define GL(a) ((CONST struct gl_list_t *)a)
/* macro to convert the gl_list to the ChildListPtr */
#define CLC(a) ((ChildListPtr)a)

int CmpChildListEntries(CONST struct ChildListEntry *e1,
                        CONST struct ChildListEntry *e2)
{
  assert(e1!=NULL);
  assert(e2!=NULL);
  assert(e1->strptr!=NULL);
  assert(e2->strptr!=NULL);
  if (e1->strptr==e2->strptr) return 0;
  return strcmp(e1->strptr,e2->strptr);
}

static
struct ChildListEntry *copylistentry(struct ChildListEntry *old)
{
  struct ChildListEntry *new;
  new = CLPOOLALLOC;
  new->strptr = old->strptr;

  new->typeptr = old->typeptr;
  CopyTypeDesc((struct TypeDescription *)old->typeptr);

  new->statement = old->statement;
  new->bflags = old->bflags;
  new->isarray = (old->isarray > 0) ? old->isarray : 0;
  new->origin = LegalOrigin(old->origin);
  return new;
}

static
int cl_inputIsOK(struct gl_list_t *l)
{
  unsigned long c,length;
  struct ChildListEntry *old;
  length = gl_length(l);
  for (c=1;c<=length;c++) {
    old = CGET(l,c);
    if (old->typeptr == NULL) {
      return 0;
    }
  }
  return 1;
}

ChildListPtr CreateChildList(struct gl_list_t *l)
{
  register struct ChildListEntry *old;
  register struct gl_list_t *result;
  register unsigned long c,length;
  assert(l!=NULL);
  length = gl_length(l);
  result = gl_create(length);
  if (!cl_inputIsOK(l)) {
    return NULL;
  }
  if (!gl_sorted(l)) gl_sort(l,(CmpFunc)CmpChildListEntries);
  for (c=1;c<=length;c++) {
    old = CGET(l,c);
    assert(old!=NULL);
    CPUT(result,CLCOPY(old)); /* uses copylistentry(old) */
  }
  return (ChildListPtr)result;
}

void DestroyChildList(ChildListPtr cl)
{
  unsigned long int c,len;
  struct ChildListEntry *old;
  assert(cl!=NULL);
  len = gl_length(GL(cl));
  for (c=1;c<=len;c++) {
    old = CGET(cl,c);
    DeleteTypeDesc((struct TypeDescription *)old->typeptr);
    old->typeptr = NULL;
    CLPOOLFREE(old);
  }
  gl_destroy((struct gl_list_t *)GL(cl));
}

ChildListPtr AppendChildList(ChildListPtr cl,
                             struct gl_list_t *l)
{
  register struct gl_list_t *result;
  struct ChildListEntry *cle, *gle;
  register unsigned long length,cln,gln;
  register unsigned long clength;
#ifndef NDEBUG
  register unsigned long totlen;
#endif
  assert(l!=NULL);
  assert(cl!=NULL);
  if (!cl_inputIsOK(l)) {
    return NULL;
  }
  length = gl_length(l);
  if (!gl_sorted(l)) gl_sort(l,(CmpFunc)CmpChildListEntries);

  clength = gl_length((CONST struct gl_list_t *)cl);
  if (!clength) return CreateChildList(l);
  if (!length) return CreateChildList((struct gl_list_t *)cl);
  result = gl_create(clength+length);
#ifndef NDEBUG
  totlen = (clength+length);
#endif

  /* merge two sorted lists */
  /*  cle = clnth element of cl.  */
  /*  gle = glnth element of gl.  */
  gln=1;
  cln=1;
  gle = gl_fetch(l,gln);
  cle = CGET(cl,cln);
  while (1) {
    /* once here always returns to caller for finite lists */
    if (CmpChildListEntries(gle,cle)>0) {
      CPUT(result,CLCOPY(cle));
      cle = CGET(cl,++cln);
      if (--clength==0) {
        /* copy tail of l */
        while (gln++ <= length) {
          CPUT(result,CLCOPY(gl_fetch(l,gln)));
        }
        assert(totlen== gl_length(result));
        return CLC(result);
      }
    } else {
      CPUT(result,CLCOPY(gle));
      if (++gln <= length) {
        gle = gl_fetch(l,gln);
      } else {
        /* copy tail of cl */
        while(clength-->0) {
          CPUT(result,CLCOPY(CGET(cl,cln)));
          cln++;
        }
        assert(totlen== gl_length(result));
        return CLC(result);
      }
    }
  }
  /* not reached */
}

unsigned long ChildListLen(ChildListPtr cl)
{
  assert(cl!=NULL); /* even childless defs are supposed to have empty list */
  return gl_length(GL(cl));
}

symchar *ChildStrPtr(ChildListPtr cl, unsigned long int n)
{
  assert(cl!=NULL && CGET(cl,n)->strptr!=NULL);
  return CGET(cl,n)->strptr;
}

unsigned int ChildIsArray(ChildListPtr cl,unsigned long int n)
{
  return (unsigned int)CGET(cl,n)->isarray;
}

unsigned int ChildOrigin(ChildListPtr cl,unsigned long int n)
{
  return CGET(cl,n)->origin;
}

unsigned int ChildParametric(ChildListPtr cl,unsigned long int n)
{
  return ParametricOrigin(CGET(cl,n)->origin);
}

CONST struct Statement *ChildStatement(ChildListPtr cl, unsigned long int n)
{
  assert(cl!=NULL && CGET(cl,n)->statement!=NULL);
  return CGET(cl,n)->statement;
}

unsigned ChildGetBooleans(ChildListPtr cl, unsigned long int n)
{
  assert(cl!=NULL && n && n <= gl_length(GL(cl)));
  return CGET(cl,n)->bflags;
}

void ChildSetBoolean(ChildListPtr cl, unsigned long int n,
                     unsigned flag, unsigned bool)
{
  assert(cl!=NULL && n && n <= gl_length(GL(cl)));
  switch (flag) {
  case CBF_VISIBLE:
  case CBF_SUPPORTED:
    break;
  default:
    return;
  }
  if (bool==1) {
    CGET(cl,n)->bflags |= flag;
  } else {
    CGET(cl,n)->bflags &= ~flag;
  }
}

CONST struct TypeDescription *ChildBaseTypePtr(ChildListPtr cl,
                                               unsigned long int n)
{
  assert(cl!=NULL);
  return CGET(cl,n)->typeptr;
}

unsigned long ChildPos(ChildListPtr cl, CONST char *s)
/*********************************************************************\
Use binary search to locate the string.  This routine will be called
alot, so it is important to make it quick.

Yah, well, it could be a helluva lot quicker if it just compared
pointers maybe. Do we have a fundamental breakdown in design here?
Are we sorting on the right things? Did we ever do profiling?
If it's so damn important, why don't we reimplement the symbol table?
\*********************************************************************/
{
#define PTRSRCHFIRST 0 /* BAA */

  register unsigned long lower,upper,search;
  register int comp;
  lower=1;
  upper=gl_length(GL(cl));

#if PTRSRCHFIRST

  /* linear search sucks when childlists are long */
  for (search = 0; search <upper;search++) {
    if (CGET(cl,search+1)==s) return (search+1);
  }
  /* if we get here, it's because string isn't in the childlist,
   * (a less usual case in a good model) and we do the nasty strcmp.
   * What we really need is a dynamically indexed symbol table so
   * that we can make symchar comparison fast.
   * e.g.
   * struct symchar {
   *   CONST char *name;
   *   unsigned long refcount;
   *   unsigned long index;
   *   struct symchar *prev;
   *   struct symchar *next;
   * };
   * The symchar should be in a hash table bucket {struct symchar *, next};
   * and also in a D.L. list with index initially assigned values
   * ~10000 apart so that we do a minimum of reindexing as we insert.
   * We should keep an array of pivots too to speed insertion.
   * Equation labels will give us hell. so clever is at a premium.
   * and the need for int comparisons is great.
   * the need for refcount is arguable.
   */
  while (lower<=upper) {
    search = (lower+upper)/2;
    comp = strcmp(CGET(cl,search)->strptr,s);
    if (comp==0) return search;
    if (comp < 0) {
      lower = search + 1;
    } else {
      upper = search - 1;
    }
  }
  return 0;

#else

  /* strcmp is a really crappy comparator when looking for relations */
  while (lower<=upper) {
    search = (lower+upper)/2;
    comp = strcmp(CGET(cl,search)->strptr,s);
    if (comp==0) return search;
    if (comp < 0) {
      lower = search + 1;
    } else {
      upper = search - 1;
    }
  }
  return 0;

#endif
}

unsigned long FastChildPos(ChildListPtr cl, symchar *s)
{
  static FastChildPos_not_implemented =0;
  (void) cl; (void) s;
  assert(FastChildPos_not_implemented);
  return 0;
}

extern int CompareChildLists(ChildListPtr cl1,
                             ChildListPtr cl2,
                             unsigned long *diff)
{
  struct ChildListEntry *cle1, *cle2;
  unsigned long int len1, len2, c, len;
  int ctmp;
  if (cl1==cl2) {
    return 0;
  }
  assert(cl1!= NULL);
  assert(cl2!= NULL);
  assert(diff != NULL);
  *diff = 0;
  len1=ChildListLen(cl1);
  len2=ChildListLen(cl2);
  /* do not return early just because len1 != len2. want to check
   * equivalency up to the length of the shorter.
   */
  len = MIN(len1,len2);
  if (len==0) {
    /* curiously  empty lists > lists with something */
    if (len1) return -1;
    if (len2) return 1;
    return 0;
  }
  for (c=1; c <= len; c++) { /* we always enter this loop */
    cle1 = CGET(cl1,c);
    cle2 = CGET(cl2,c);
    if (cle1->strptr != cle2->strptr) {
      ctmp = CmpSymchar(cle1->strptr , cle2->strptr);
      assert(ctmp);
      break;
    }
    if (cle1->typeptr != cle2->typeptr) {
      ctmp = (GetParseId(cle1->typeptr) > GetParseId(cle2->typeptr) ? 1 : -1);
      break;
    }
    if (cle1->bflags != cle2->bflags) {
      ctmp = ((cle1->bflags > cle2->bflags) ? 1 : -1);
      break;
    }
    if (cle1->origin != cle2->origin) {
      ctmp = ((cle1->origin > cle2->origin) ? 1 : -1);
      break;
    }
    if (cle1->isarray != cle2->isarray) {
      ctmp = ((cle1->isarray > cle2->isarray) ? 1 : -1);
      break;
    }
    /* passing bflags, origin, and isarray tests pretty much assures
     * passing statement comparison test, too.
     */
  }
  *diff = c;
  if (c <= len) {
    /* finished before end of short list */
    return ctmp;
  }
  if (len1 == len2) {
    /* same len. finished both lists */
    *diff = 0;
    return 0;
  }
  if (len > len2) {
    /* identical up to len. list length decides */
    return 1;
  } else {
    return -1;
  }
}

void WriteChildList(FILE *fp,ChildListPtr cl)
{
  unsigned long c,len;
  struct ChildListEntry *cle;
  CONST struct gl_list_t *l;
  if (cl!=NULL) {
    l = GL(cl);
    len = gl_length(l);
    if (!len) {
      FPRINTF(fp,"Child list is empty\n");
      return;
    }
    FPRINTF(fp,"Child list is %lu long.\n",len);
    for (c=1;c<=len;c++) {
      cle = CGET(l,c);
      if (cle!=NULL) {
        FPRINTF(fp,"%lu name=\"%s\" type=\"%s\" ARRAY=%d origin=%d\n",
          c,
          cle->strptr,
          ((cle->typeptr==NULL)?"UNKNOWN":GetName(cle->typeptr)),
          cle->isarray,
          (int)cle->origin);
        WSEM(fp,cle->statement,"  Declared at ");
      } else {
        FPRINTF(fp,"Child list item %lu is empty!\n",c);
      }
    }
  }
}
