/*
 *  Symbol Table Management
 *  by Tom Epperly
 *  Version: $Revision: 1.6 $
 *  Version control file: $RCSfile: symtab.c,v $
 *  Date last modified: $Date: 97/07/18 11:41:04 $
 *  Last modified by: $Author: mthomas $
 *
 *  This file is part of the Ascend Language Interpreter.
 *
 *  Copyright (C) 1990, 1993, 1994 Thomas Guthrie Epperly
 *
 *  The Ascend Language Interpreter is free software; you can redistribute
 *  it and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  The Ascend Language Interpreter is distributed in hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge, MA 02139 USA.  Check the file named
 *  COPYING.
 *
 */
#include <stdarg.h>
#include "utilities/ascConfig.h"
#include "compiler/compiler.h"
#include "utilities/ascPanic.h"
#include "utilities/ascMalloc.h"
#include "compiler/symtab.h"
#include "general/hashpjw.h"

/*
 *  For various reasons which are still unclear to me, it is best to chose
 *  a symbol table size which is a prime number.
 */
#define SYMTABSIZE 1021

#ifndef lint
static CONST char SymbolTableID[] = "$Id: symtab.c,v 1.6 97/07/18 11:41:04 mthomas Exp $";
#endif

struct StringSpaceRec {
  struct StringSpaceRec *next;
  unsigned long used;
  char data[MAXIMUM_STRING_LENGTH];
};

struct symbol_entry {
  CONST char *entry;
  struct symbol_entry *next;
};

/*********************************************************************\
Global Variables
\*********************************************************************/
static struct StringSpaceRec *g_string_space=NULL;
struct symbol_entry *g_symbol_table[SYMTABSIZE];
static unsigned long g_symbol_size = 0;
static unsigned long g_symbol_collisions = 0;


/*********************************************************************\
  String Space Management. Support routines for the symbol table.
\*********************************************************************/

/*********************************************************************\
const char *CopyString(str)
const char *str;
Make a copy of "str" in the string space.
\*********************************************************************/
static symchar *CopyString(CONST char *str,int userlen)
{
  register unsigned long length;
  register struct StringSpaceRec *ptr;
  register char *result;
  length = (unsigned long)userlen+1;
  assert(length<=MAXIMUM_STRING_LENGTH);
  if ((ptr = g_string_space)!=NULL){
    while ((ptr != NULL)&&((length+ptr->used)>=MAXIMUM_STRING_LENGTH))
      ptr = ptr->next;
    if (ptr!=NULL){
      result = &ptr->data[ptr->used];
      strcpy(result,str);
      ptr->used += length;
      return result;
    }
    ptr = g_string_space;
  }
  if ((g_string_space =
       (struct StringSpaceRec *)ascmalloc(sizeof(struct StringSpaceRec)))
      ==NULL){
    Asc_Panic(2, NULL, "Unable to allocate string space.\n");
    exit(2);/* Needed to keep gcc from whining */
  }
  g_string_space->next = ptr;
  g_string_space->used = length;
  strcpy(g_string_space->data,str);
  return g_string_space->data;
}

void DestroyStringSpace(void)
{
  register struct StringSpaceRec *ptr,*next;
  ptr = g_string_space;
  while (ptr != NULL){
    next = ptr->next;
#ifndef NDEBUG
    ascbzero((char *)ptr,(int)sizeof(struct StringSpaceRec));
#endif
    ascfree((char *)ptr);
    ptr = next;
  }
  g_string_space = NULL;
}

static void StringSpaceReport(void)
{
  register unsigned long num_blocks=0,used_memory=0;
  register struct StringSpaceRec *ptr;
  ptr = g_string_space;
  while(ptr != NULL){
    num_blocks++;
    used_memory += ptr->used;
    ptr = ptr->next;
  }
  PRINTF("Number of blocks : %lu\n",num_blocks);
  PRINTF("Total Memory use : %lu\n",num_blocks*sizeof(struct StringSpaceRec));
  PRINTF("String memory use: %lu\n",num_blocks*MAXIMUM_STRING_LENGTH);
  PRINTF("Actual usage     : %lu\n",used_memory);
  PRINTF("Fraction filled  : %f\n",(double)used_memory/
	 ((double)num_blocks*(double)sizeof(struct StringSpaceRec)));
}


/*********************************************************************\
  Real Symbol Table Routines
\*********************************************************************/

void InitSymbolTable(void)
{
  register int c;
  for(c=0;c<SYMTABSIZE;g_symbol_table[c++]=NULL);
  g_symbol_size = 0;
  g_symbol_collisions = 0;
}


#define SymHashFunction(s) hashpjw((s),SYMTABSIZE)

symchar *AddSymbolL(CONST char *c, int l)
{
  register unsigned long hv;
  register int cmp;
  register struct symbol_entry *item,*next;
  hv = SymHashFunction(c);
  if (l<=0) l = strlen(c);
  if (g_symbol_table[hv]==NULL) {
    item =
      (struct symbol_entry *)ascmalloc((unsigned)sizeof(struct symbol_entry));
    item->next = NULL;
    item->entry = CopyString(c,l);
    g_symbol_table[hv] = item;
    g_symbol_size++;
  }
  else {
    item = g_symbol_table[hv];
    if ((cmp = strcmp(item->entry,c))>0) {
      item =
	(struct symbol_entry *)ascmalloc((unsigned)sizeof(struct symbol_entry));
      item->next = g_symbol_table[hv];
      item->entry = CopyString(c,l);
      g_symbol_table[hv] = item;
      g_symbol_size++;
      g_symbol_collisions++;
    }
    else if (cmp<0) {
      for(next=item->next;next!=NULL;item=next,next=next->next) {
	if ((cmp=strcmp(next->entry,c))>=0) {
	  if (cmp==0) return next->entry;
	  break;
	}
      }
      item->next = (struct symbol_entry *)
	ascmalloc((unsigned)sizeof(struct symbol_entry));
      item = item->next;
      item->next = next;
      item->entry = CopyString(c,l);
      g_symbol_size++;
      g_symbol_collisions++;
    }
  }
  return item->entry;
}

symchar *AddSymbol(CONST char *c)
{
  return AddSymbolL(c,strlen(c));
}

void PrintTab(int noisy)
{
  unsigned long c,counts[SYMTABSIZE];
  struct symbol_entry *item;
  for(c=0;c<SYMTABSIZE;counts[c++]=0);
  PRINTF("Symbol table contains %lu entries with %lu collisions.\n",
	 g_symbol_size,g_symbol_collisions);
  if (noisy) {
    for(c=0;c<SYMTABSIZE;c++) {
      for(item=g_symbol_table[c];item!=NULL;item=item->next) {
        PRINTF("%lu %s\n",c,item->entry);
        counts[c]++;
      }
    }
    PRINTF("Distribution information\n");
    for(c=0;c<SYMTABSIZE;c++) PRINTF("%lu\t%lu\n",c,counts[c]);
  }
  StringSpaceReport();
}

void DestroySymbolTable(void)
{
  unsigned int c;
  struct symbol_entry *item,*next;
  for(c=0;c<SYMTABSIZE;g_symbol_table[c++]=NULL) {
    for(item=g_symbol_table[c];item!=NULL;item=next) {
      next = item->next;
      ascfree((char *)item);
    }
  }
}
