/*
 *  Ascend Instance Tree Type Implementation
 *  by Tom Epperly
 *  9/3/89
 *  Version: $Revision: 1.8 $
 *  Version control file: $RCSfile: tmpnum.c,v $
 *  Date last modified: $Date: 97/09/08 18:08:26 $
 *  Last modified by: $Author: ballan $
 *
 *  This file is part of the Ascend Language Interpreter.
 *
 *  Copyright (C) 1990, 1993, 1994 Thomas Guthrie Epperly
 *
 *  The Ascend Language Interpreter is free software; you can redistribute
 *  it and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  The Ascend Language Interpreter is distributed in hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge, MA 02139 USA.  Check the file named
 *  COPYING.
 *
 */
#include <stdarg.h>
#include "utilities/ascConfig.h"
#include "utilities/ascPanic.h"
#include "utilities/ascMalloc.h"
#include "general/pool.h"
#include "general/list.h"
#include "general/dstring.h"
#include "compiler/compiler.h"
#include "compiler/bit.h"
#include "compiler/symtab.h"
#include "compiler/fractions.h"
#include "compiler/dimen.h"
#include "compiler/functype.h"
#include "compiler/types.h"
#include "compiler/childinfo.h"
#include "compiler/child.h"
#include "compiler/type_desc.h"
#include "compiler/instance_enum.h"
#include "compiler/instance_name.h"
#include "compiler/instance_io.h"
#include "compiler/instance_types.h"
#include "compiler/instmacro.h"
#include "compiler/visitinst.h"
#include "compiler/extinst.h"
#include "compiler/instquery.h"
#include "compiler/linkinst.h"
#include "compiler/destroyinst.h"
#include "compiler/createinst.h"
#include "compiler/refineinst.h"
#include "compiler/atomvalue.h"
#include "compiler/atomsize.h"
#include "compiler/check.h"
#include "compiler/dump.h"
#include "compiler/prototype.h"
#include "compiler/pending.h"
#include "compiler/find.h"
#include "compiler/relation_type.h"
#include "compiler/relation.h"
#include "compiler/logical_relation.h"
#include "compiler/logrelation.h"
#include "compiler/relation_util.h"
#include "compiler/logrel_util.h"
#include "compiler/rel_common.h"
#include "compiler/case.h"
#include "compiler/when_util.h"
#include "compiler/universal.h"
#include "compiler/cmpfunc.h"
#include "compiler/tmpnum.h"

#ifndef lint
static CONST char TmpNumModuleID[] = "$Id: tmpnum.c,v 1.8 97/09/08 18:08:26 ballan Exp $";
#endif

/*
 * Modified by kaa 12 Sep 1995, to be more useful friendly
 * and more efficient. Does not choke on fundamentals.
 * It simply returns an invalid result; 0 is almost never
 * a valid result in the compiler.
 */
unsigned long GetTmpNum(CONST struct Instance *i)
{
  AssertMemory(i);
  if (i == NULL) {
    return LONG_MAX; /* baa */
  }
  if (IsAtomicInstance(i)) {
    return CA_INST(i)->tmp_num;
  }
  switch (i->t) {
  case MODEL_INST:
    return MOD_INST(i)->tmp_num;
  case REL_INST:
    return RELN_INST(i)->tmp_num;
  case LREL_INST:
    return LRELN_INST(i)->tmp_num;
  case WHEN_INST:
    return W_INST(i)->tmp_num;
  case ARRAY_INT_INST:
  case ARRAY_ENUM_INST:
    return ARY_INST(i)->tmp_num;
  case REAL_CONSTANT_INST:
    return RC_INST(i)->tmp_num;
  case BOOLEAN_CONSTANT_INST:
    return BC_INST(i)->tmp_num;
  case INTEGER_CONSTANT_INST:
    return IC_INST(i)->tmp_num;
  case SYMBOL_CONSTANT_INST:
    return SYMC_INST(i)->tmp_num;
  case SIM_INST:
    return SIM_INST(i)->tmp_num;
  case DUMMY_INST:
    return D_INST(i)->tmp_num;
  case REAL_INST:		/* fall through */
  case INTEGER_INST:		/* fall through */
  case BOOLEAN_INST:		/* fall through */
  case SET_INST:		/* fall through */
  case SYMBOL_INST:
    return 0;		/* this is the change was exit(2).kaa */
  default:
    Asc_Panic(2, NULL, "Incorrect type in GetTmpNum.\n");
    return LONG_MAX; /* but not really. shuts up gcc, though. */
  }
}

void SetTmpNum(struct Instance *i, unsigned long int cn)
{
  AssertMemory(i);
  if (IsAtomicInstance(i)) {
    CA_INST(i)->tmp_num = cn;
    return;
  }
  if (IsConstantInstance(i)) {
    CI_INST(i)->tmp_num = cn;
    return;
  }
  switch(i->t) {
  case SIM_INST:
    SIM_INST(i)->tmp_num = cn;
  case DUMMY_INST:
    D_INST(i)->tmp_num = cn;
    break;
  case MODEL_INST:
    MOD_INST(i)->tmp_num = cn;
    break;
  case REL_INST:
    RELN_INST(i)->tmp_num = cn;
    break;
  case LREL_INST:
    LRELN_INST(i)->tmp_num = cn;
    break;
  case WHEN_INST:
    W_INST(i)->tmp_num = cn;
    break;
  case ARRAY_INT_INST:
  case ARRAY_ENUM_INST:
    ARY_INST(i)->tmp_num = cn;
    break;
  case REAL_INST:		/* fall through */
  case INTEGER_INST:		/* fall through */
  case BOOLEAN_INST:		/* fall through */
  case SET_INST:		/* fall through */
  case SYMBOL_INST:
    break;		/* another change -- dont choke just return */
  default:
    Asc_Panic(2, NULL, "Incorrect type in SetTmpNum.\n");
  }
}

static void ResetNodes(struct Instance *i)
{
  SetTmpNum(i,0);
}

void ZeroTmpNums(struct Instance *i,int order)
{
    SilentVisitInstanceTree(i,ResetNodes,order,0);
}

