/*
 *  Problem Analysis Routines
 *  by Benjamin Andrew Allan
 *  5/19/96
 *  Version: $Revision: 1.42 $
 *  Version control file: $RCSfile: analyze.c,v $
 *  Date last modified: $Date: 97/10/04 19:34:40 $
 *  Last modified by: $Author: rv2a $
 *  Copyright(C) 1996 Benjamin Andrew Allan
 *
 *  This file is part of the ASCEND IV math programming system.
 *
 *  The SLV solver is free software; you can redistribute
 *  it and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  The SLV solver is distributed in hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge, MA 02139 USA.  Check the file named
 *  COPYING.  COPYING is found in ../compiler.
 *
 *
 *  These functions are the start of a new design for feeding
 *  solvers from the ASCEND compiler or any arbitrary backend.
 *
 *  The intention is that eventually the other code in the solver
 *  directory will really be for *solution algorithms* and the
 *  definition of a *problem* will come from here. In essence, most
 *  of what was solver/system.c will be here. Negotiating directly
 *  with the ASCEND instance hierarchy should not be a solver's
 *  job.
 *  The goal of this module is to CREATE a slv_system_t data structure
 *  capable of supporting code generation, an interactive interface, and
 *  in-core solvers, while being expandable in the future to out of core
 *  solvers/external-process solvers.
 *
 *  A secondary goal is to have nonlinear solver files be independent of
 *  all the compiler directory files except ascmalloc.h.
 *  The present fly in the ointment is expr.h because of the objective fcns.
 *  The relman and exprman modules go away because they are indicative of
 *  functionality that belongs either in the compiler or rel.c.
 *  If we meet this goal, then it is a simple matter to connect any
 *  arbitrary compiler backend to the solver API by replacing the rel
 *  and var and analyze modules.
 *
 */

#include <stdarg.h>
#include "utilities/ascConfig.h"
#include "utilities/ascPanic.h"
#include "utilities/ascMalloc.h"
#include "general/list.h"
#include "general/dstring.h"
#include "compiler/instance_enum.h"
#include "compiler/fractions.h"
#include "compiler/compiler.h"
#include "compiler/dimen.h"
#include "compiler/atomvalue.h"
#include "compiler/parentchild.h"
#include "compiler/visitinst.h"
#include "compiler/types.h"
#include "compiler/exprs.h"
#include "compiler/sets.h"
#include "compiler/mathinst.h"
#include "compiler/instquery.h"
#include "compiler/instance_io.h"
#include "compiler/relation_type.h"
#include "compiler/find.h"
#include "compiler/extfunc.h"
#include "compiler/extcall.h"
#include "compiler/relation.h"
#include "compiler/functype.h"
#include "compiler/safe.h"
#include "compiler/relation_util.h"
#include "compiler/logical_relation.h"
#include "compiler/logrelation.h"
#include "compiler/logrel_util.h"
#include "compiler/case.h"
#include "compiler/when_util.h"
#include "solver/mtx.h"
#include "solver/slv_types.h"
#include "solver/var.h"
#define _SLV_SERVER_C_SEEN_
#include "solver/rel.h"
#include "solver/logrel.h"
#include "solver/discrete.h"
#include "solver/conditional.h"
#include "solver/bnd.h"
#include "solver/slv_server.h"
#include "solver/slv_common.h"
#include "solver/linsol.h"
#include "solver/linsolqr.h"
#include "solver/slv_client.h"
#include "solver/cond_config.h"
#include "solver/analyze.h"

/* stuff to get rid of */
#ifndef MAX_VAR_IN_LIST
#define MAX_VAR_IN_LIST 20
#endif /* MAX_VAR_IN_LIST  */

/* 
 * Forward declaration 
 */
static void ProcessModelsInWhens(struct Instance *, struct gl_list_t *,
                                 struct gl_list_t *, struct gl_list_t *);

/* 
 * Global variable. Set to true by classify if need be
 */
static int g_bad_rel_in_list;


/* used to give an integer value to each symbol used in a when */
struct gl_list_t *g_symbol_values_list = NULL;

struct varip {
  struct var_variable *data;  /* ptr to destination of data */
  int index;		      /* master gl index */
  int incident;		      /* set 0 in classify_instance,
				 1 make_master_lists */
  int in_block;		      /* set 0 in classify_instance */
  int fixed;		      /* set in classify_instance */
  int solvervar;	      /* set in classify_instance */
  int active;                 /* is this var a part of my problem */
};


struct disvarip {
  struct dis_discrete *data;	/* ptr to destination of data */
  int index;		        /* master gl index */
  int fixed;		        /* set in classify_instance */
  int isconst;                  /* is this dis var constant ? */
  int distype;                  /* 0 boolean, 1 int, -1 symbol */
  int value;                    /* integer value of the variable */
  int incident;                 /* is it incident in a logrel */
  int inwhen;                   /* is it in a when var list */
  int booleanvar;               /* Not sure if I need it */
  int active;                   /* is this disvar a part of my problem */
};


struct relip {
  struct rel_relation *data;    /* ptr to destination of data */
  long model;		/* relation is in this model in model gllist */
			/* set in CollectRelsAndWhens. = 1.. nmodels */
                        /* rel_relation models = u.r.model-1 */
  int index; 		/* master gl list index */
  int obj;	/* is it an objective relation. set in classify_instance */
  int ext;	/* is it e_blackbox. set in classify_instance */
  int included; /* set in classify_instance */
  int cond;	/* is it a conditional relation. set in classify_instance */
  int inwhen;   /* is it in a when */
  int active;   /* is this rel a part of my problem */
};

struct logrelip {
  struct logrel_relation *data;    /* ptr to destination of data */
  long model;		/* logrelation is in this model in model gllist */
  int index; 		/* master gllist index */
  int included;         /* set in classify_instance */
  int cond;	/* is it a conditional logrelation.  */
  int inwhen;   /* is it in a when */
  int active;   /* is this logrel a part of my problem */
};

struct whenip{
  struct w_when *data;   /* ptr to destination of data */
  long model;	         /* when is in this model in model gllist */
  int index; 		 /* master gllist index */
  int inwhen;            /* is it in a when */
};

struct modip {
  int index;		/* set in make master lists. 1..nmodels */
  int inwhen;           /* is it in a when */
};

/* we will decorate the ascend instance tree with these in the interface
 * pointers, but ONLY for the system build process and not persistently.
 * DO NOT EVER UNDER ANY CIRCUMSTANCES EXPORT THIS DATA STRUCTURE.
 */
struct solver_ipdata {
  struct Instance *i; /* the kind of instance is the enum for union */
  union {
    struct modip m;
    struct varip v;
    struct disvarip dv;
    struct relip r;
    struct logrelip lr;
    struct whenip w;
  } u;
};

/* a handy cast for fetching things off gllists */
#define SIP(x) ((struct solver_ipdata *)(x))

/* 
 * a bridge buffer used so much we aren't going to free it, just reuse it 
 */
static struct reuse_t {
  size_t ipcap;			/* number of ips allocated in ipbuf */
  size_t ipused;		/* number of ips in use */
  struct solver_ipdata *ipbuf;
} g_reuse = {0,0,NULL};

/* 
 *  a data structure for bridge building only. hell of a scaffolding.
 *  all fields should be empty if construction is not in progress. 
 */
static struct problem_t {
/* the following are established by CountStuffInTree */
  long nv;		/* number of solvervar/solveratom */
  long np;		/* number of real ATOM instance parameters */
  long nu;		/* number of real ATOM instance uninteresting */
  long ndv;             /* number of discrete variables */
  long nud;             /* number of uninteresting discretes */
  long nc;		/* number of conditional relations */
  long ncl;		/* number of conditional logrelations */
  long nr;		/* number of algebraic relations */
  long no; 		/* number of objective rels */
  long nl;		/* number of logical rels */
  long nw;              /* number of whens */
  long ne;		/* number of external rels subset overestimate*/
  long nm;		/* number of models */
/* 
 * The following gllists contain pointers to interface ptrs as locally defined.
 * The lists will be in order found by a visit instance tree.
 */
  struct gl_list_t *vars;	/* solvervar/solveratom. varips */
  struct gl_list_t *pars;	/* real ATOM instance parameters */
  struct gl_list_t *unas;	/* real ATOM instance of no 'apparent' use */
  struct gl_list_t *models;	/* models in tree. modips */
/* 
 * The following gllists contain pointers to interface ptrs as locally defined.
 * The lists will be in order found by running over the models list.
 */
  struct gl_list_t *dvars;	/* discrete variables */
  struct gl_list_t *dunas;	/* discrete variables of no use */
  struct gl_list_t *whens;	/* whens */
  struct gl_list_t *cnds;	/* conditional relations */
  struct gl_list_t *logcnds;	/* conditional logrelations */
  struct gl_list_t *rels;	/* ascend relations. relips */
  struct gl_list_t *objrels;	/* objective rels. relips */
  struct gl_list_t *logrels;	/* logical rels */
/* bridge ip data */
  struct gl_list_t *oldips;	/* buffer of oldip crap we're protecting */
/* misc stuff */
  struct gl_list_t *tmplist;	/* sort space */
/* stuff that will end up in the slv_system_t */
  struct rel_relation *obj;	       /* DEFAULT objective relation, if any */
  struct Instance *root;	/* instance we construct system from */
  struct gl_list_t *extrels;	/* black box stub list */
/* stuff that should move elsewhere, but end up in slv_system_t */
  mtx_region_t *blocks;		/* array of partitions in reordered matrix */
  int32 nblocks;		/* size of array of partitions */
  int nnz;			/* free nonzeros in processed jacobian */
  int nnztot;			/* total nonzeros in processed relations */
  int nnzobj;			/* total nonzeros in objective gradients */
  int nnzcond;			/* total nonzeros in conditional relations */
  int relincsize;		/* total nonzeros in gradients */
  int relincinuse;		/* incidence given to relations so far */
  int varincsize;		/* total nonzeros in gradients (redundant) */
  int varincinuse;		/* incidence given to variables so far */
  int ncol;			/* free and incident vars */
  int nrow;			/* included relations */
 /* logical relation stuff */
  int lognnz;			/* Summ of free boolean vars in inc logrels */
  int lognrow;			/* included logrelations */
  int logncol;			/* free and incident boolean vars */
  int lrelinc;                  /* incident boolean vars */
  int lrelincsize;	        /* Total summ of incidences (boolean vars)
				   in logrels*/
  int lrelincinuse;		/* incidence given to log relations so far */
/* data to go to slv_system_t */
  struct rel_relation *reldata;      /* rel data space, mass allocated */
  struct rel_relation *objdata;      /* objrel data space, mass allocated */
  struct rel_relation *condata;      /* cond rel data space, mass allocated */
  struct logrel_relation *lrdata;    /* logrel data space, mass allocated */
  struct logrel_relation *logcondata; /* cond logrel data space, allocated */
  struct var_variable *vardata;      /* svar data space, mass allocated */
  struct var_variable *pardata;      /* par data space, mass allocated */
  struct var_variable *undata;       /*  data space, mass allocated */
  struct dis_discrete *disdata;      /* dis var data space, mass allocated */
  struct dis_discrete *undisdata;    /*  data space, mass allocated */
  struct w_when *whendata;           /* when data space, mass allocated */
  struct bnd_boundary *bnddata;      /* boundaries data space, allocated */
  struct var_variable **mastervl;	/* master null-terminated list */
  struct var_variable **solvervl;	/* solvers null-terminated list */
  struct dis_discrete **masterdl;	/* master null-terminated list */
  struct dis_discrete **solverdl;	/* solvers null-terminated list */
  struct rel_relation **masterrl;	/* master null-terminated list */
  struct rel_relation **solverrl;	/* solvers null-terminated list */
  struct rel_relation **mastercl;	/* master null-terminated list */
  struct rel_relation **solvercl;	/* solvers null-terminated list */
  struct rel_relation **masterol;	/* master null-terminated list */
  struct rel_relation **solverol;	/* solvers null-terminated list */
  struct logrel_relation **masterll;	/* master null-terminated list */
  struct logrel_relation **solverll;	/* solvers null-terminated list */
  struct logrel_relation **mastercll;	/* master null-terminated list */
  struct logrel_relation **solvercll;	/* solvers null-terminated list */
  struct w_when **masterwl;     	/* master null-terminated list */
  struct w_when **solverwl;	        /* solvers null-terminated list */
  struct bnd_boundary **masterbl;     	/* master null-terminated list */
  struct bnd_boundary **solverbl;	/* solvers null-terminated list */
  struct var_variable **masterpl;	/* master null-terminated list */
  struct var_variable **solverpl;	/* solvers null-terminated list */
  struct var_variable **masterul;	/* master null-terminated list */
  struct var_variable **solverul;	/* solvers null-terminated list */
  struct dis_discrete **masterdul;	/* master null-terminated list */
  struct dis_discrete **solverdul;	/* solvers null-terminated list */
  struct var_variable **relincidence;	/* rel_relation incidence source */
  struct rel_relation **varincidence;	/* var_variable incidence source */
  struct dis_discrete **logrelinciden;	/* logrel_relation incidence source */

  struct ExtRelCache **erlist;	/* external rel cache null terminated list */
} g_prob;
/* we are making the ANSI assumption that this will be init to 0/NULL*/
/*
 * container for globals during assembly.
 * At present, the mastervl and solvervl are of the same length. This
 * is purely coincidental and the long run intent is that there is one
 * master list and that a problem coordinator would set up the
 * solver var/rel lists blockwise as we go along. We may want to put
 * block information in the rel/var structures.
 */


/*
 * The intent here is to do away with the old persistent interface pointer
 * scheme by making the struct rel_relation *individually keep track of the
 * map between the ascend RelationVariable list position and the
 * solver's var list index (and hence the column in jacobian for jacobian
 * involved clients).
 * In this mapping each struct relation * has its var list and this list
 * may contain RealAtomInstances that we don't consider variables.
 * In the rel_relation we will have the variable index list
 * which is an array of int32 the same length as the RelationVariable list.
 * In the array position 0 corresponds to RelationVariable 1 since the
 * compiler uses gl_lists. If in the variable index list we encounter
 * a number < 0 we know that that RelationVariable doesn't map to what
 * we consider a solver variable.
 *
 * In the near future we may also add to the struct rel_relation *an array,
 * a, of int32 pairs like so:
 * vlindex | rvindex | vlindex | rvindex
 * and a length. This array could be built of the data for vars that pass
 * a filter provided by the client. This way a client could help us avoid
 * having to do if testing while stuffing jacobians.
 * In this scheme stuffing a jacobian row (or whatever) would simply mean
 * calling the compiler's derivative function (wrt RelationVariable list)
 * which returns a vector d of values and then doing a loop:
 *   for( i = 0 ; i < length; i++) { coord.row fixed already
 *     coord.col = a[i++];
 *     mtx_fill_org_value(mtx,&coord,d[a[i]])
 *   }
 * }
 *
 * One begins to wonder if there isn't a better way to do all this, but
 * so far nothing has occurred.
 * The best way would be to feed clients only the block of stuff they
 * are interested in (no fixed, unattached vars or unincluded rels
 * or vars/rels solved in previous partitions) so that the solver
 * had no partitioning or var classification work to do except perhaps
 * classifying things as basic/nonbasic.
 */


/* return a pointer to the oncesizefitsall ips we're using.
 * always returns nonnull because if we run out we exit
 */
static struct solver_ipdata *analyze_getip(void)
{
  if (g_reuse.ipbuf!=NULL && g_reuse.ipused < g_reuse.ipcap) {
    return &(g_reuse.ipbuf[g_reuse.ipused++]);
  } else {
    Asc_Panic(2, "ananlyze_getip",
              "Too many ips requested by analyze_getip\n");
    exit(2);/* NOT REACHED.  Needed to keep gcc from whining */
  }
}

/* 
 * reallocates to the requested size (newcap) the ipbuf.
 * if newcap = 0, frees ipbuf.
 * if insufficient memory returns 1.
 * if newcap > 0 and reset mem != 0, initializes ipbuf to 0s.
 * Resets ipused to 0.
 */
static
int resize_ipbuf(size_t newcap, int resetmem)
{
  struct solver_ipdata *tmp;
  if (newcap ==0) {
    if (g_reuse.ipbuf != NULL) {
      ascfree(g_reuse.ipbuf);
    }
    g_reuse.ipcap = g_reuse.ipused = 0;
    g_reuse.ipbuf = NULL;
    return 0;
  }
  if (newcap > g_reuse.ipcap) {
    if ( g_reuse.ipcap >0 ) {
      tmp = (struct solver_ipdata *)
        ascrealloc(g_reuse.ipbuf,newcap*sizeof(struct solver_ipdata));
      if (tmp == NULL) {
        FPRINTF(ASCERR,"ERROR: (resize_ipbuf) Insufficient memory.\n");
        return 1;
      }
      g_reuse.ipbuf = tmp;
      g_reuse.ipcap = newcap;
    } else {
      tmp = (struct solver_ipdata *)
        ascmalloc(newcap*sizeof(struct solver_ipdata));
      if (tmp == NULL) {
        FPRINTF(ASCERR,"ERROR:(resize_ipbuf 1st call) Insufficient memory.\n");
        return 1;
      }
      g_reuse.ipbuf = tmp;
      g_reuse.ipcap = newcap;
    }
  }
  if (resetmem) {
    memset((char *)g_reuse.ipbuf,0,g_reuse.ipcap*sizeof(struct solver_ipdata));
  }
  g_reuse.ipused = 0;
  return 0;
}


/* 
 * checks size of request and returns a pointer to the next available
 * chunk of incidence space. if too much requested or 0 requested returns
 * NULL. g_prob.relincidence must have been allocated for this to work.
 */
static
struct var_variable **get_incidence_space(int len)
{
  struct var_variable **tmp;
  if (g_prob.relincidence == NULL) {
    Asc_Panic(2,NULL,"get_incidence_space called prematurely. bye.\n");
  }
  if (len <1) return NULL;
  if (g_prob.relincinuse + len > g_prob.relincsize) {
    FPRINTF(ASCERR,"get_incidence_space called excessively.\n");
    return NULL;
  }
  tmp = &(g_prob.relincidence[g_prob.relincinuse]);
  g_prob.relincinuse += len;
  return tmp;
}

/* 
 * checks size of request and returns a pointer to the next available
 * chunk of incidence space. if too much requested or 0 requested returns
 * NULL. g_prob.varincidence must have been allocated for this to work.
 */
static
struct rel_relation **get_var_incidence_space(int len)
{
  struct rel_relation **tmp;
  if (g_prob.varincidence == NULL) {
    Asc_Panic(2,NULL,"get_var_incidence_space called prematurely. bye.\n");
  }
  if (len <1) return NULL;
  if (g_prob.varincinuse + len > g_prob.varincsize) {
    FPRINTF(ASCERR,"get_var_incidence_space called excessively.\n");
    return NULL;
  }
  tmp = &(g_prob.varincidence[g_prob.varincinuse]);
  g_prob.varincinuse += len;
  return tmp;
}


/*
 * g_prob.logrelinciden must have been allocated for this to work.
 */
static
struct dis_discrete **get_logincidence_space(int len)
{
  struct dis_discrete **tmp;
  if (g_prob.logrelinciden == NULL) {
    Asc_Panic(2,NULL,"get_logincidence_space called prematurely. bye.\n");
  }
  if (len <1) return NULL;
  if (g_prob.lrelincinuse + len > g_prob.lrelincsize) {
    FPRINTF(ASCERR,"get_logincidence_space called excessively.\n");
    return NULL;
  }
  tmp = &(g_prob.logrelinciden[g_prob.lrelincinuse]);
  g_prob.lrelincinuse += len;
  return tmp;
}


/*
 * InitTreeCounts(i); This resets the pointers and counters in
 * g_prob to null. g_prob is supposed to be a temporary structure
 * so the memory management of the pointers in g_prob is the job
 * of the caller.
 * g_prob.root is set to i.
 */
static void InitTreeCounts(struct Instance *i)
{
  memset((char *)&g_prob,0,sizeof(struct problem_t));
  g_prob.root = i;
}

#define AVG_PARENTS 2
#define AVG_CHILDREN 4
#define AVG_RELATIONS 15
#define AVG_GROWTH 2
#define PART_THRESHOLD 1000



/*
 * The following function should be moved out to the compiler
 * under the guise of a supported attribute.
 */
static int BooleanChildValue(struct Instance *i,char *sc)
{
  if (i == NULL || sc == NULL || (i=ChildByChar(i,sc)) == NULL) {
    return 0;
  } else {
    return ( GetBooleanAtomValue(i) );
  }
}

static void CollectArrayRelsAndWhens(struct Instance *i, long modindex)
{
  struct Instance *child;
  struct solver_ipdata *rip;
  unsigned long nch,c;

  if (i==NULL) return;
  nch = NumberChildren(i);
  for (c=1;c<=nch;c++) {
    child = InstanceChild(i,c);
    if (child==NULL) continue;
    switch (InstanceKind(child)) {
    case REL_INST:
      rip = SIP(GetInterfacePtr(child));
      if (rip->u.r.model == 0) {
        rip->u.r.model = modindex;
      } else {
        /* ah hah! found an array with two distinct MODEL parents! skip it */
        break;
      }
      if (rip->u.r.cond) {
        gl_append_ptr(g_prob.cnds,(VOIDPTR)rip);
      }
      else {
        if (rip->u.r.obj) {
          gl_append_ptr(g_prob.objrels,(VOIDPTR)rip);
        } else {
          gl_append_ptr(g_prob.rels,(VOIDPTR)rip);
        }
      }
      break;
    case LREL_INST:
      rip = SIP(GetInterfacePtr(child));
      if (rip->u.lr.model == 0) {
        rip->u.lr.model = modindex;
      } else {
        break;
      }
      if (rip->u.lr.cond) {
        gl_append_ptr(g_prob.logcnds,(VOIDPTR)rip);
      }
      else {
        gl_append_ptr(g_prob.logrels,(VOIDPTR)rip);
      }
      break;
    case WHEN_INST:
      rip = SIP(GetInterfacePtr(child));
      if (rip->u.w.model == 0) {
        rip->u.w.model = modindex;
      } else {
        break;
      }
      gl_append_ptr(g_prob.whens,(VOIDPTR)rip);
      break;
    case ARRAY_ENUM_INST:
    case ARRAY_INT_INST:
      if (ArrayIsRelation(child)|| ArrayIsLogRel(child)|| ArrayIsWhen(child)) {
	CollectArrayRelsAndWhens(child,modindex);
      }
      break;
    default:
      break;
    }
  }
}


/*
 * Collect all the logrels/relations at the local scope of the MODEL
 * associated with ip->i. Local scope includes arrays of logrels/relations,
 * mainly because these arrays don't have interface pointers so we
 * can't treat them separately.
 */
static void CollectRelsAndWhens(struct solver_ipdata *ip, long modindex)
{
  struct Instance *child;
  struct solver_ipdata *rip;
  unsigned long nch,c;

  if (ip->i==NULL) return;
  nch = NumberChildren(ip->i);
  for (c=1;c<=nch;c++) {
    child = InstanceChild(ip->i,c);
    if (child==NULL) continue;
    switch (InstanceKind(child)) {
    case REL_INST:
      rip = SIP(GetInterfacePtr(child));
      rip->u.r.model = modindex;
      if (rip->u.r.cond) {
        gl_append_ptr(g_prob.cnds,(VOIDPTR)rip);
      }
      else {
        if (rip->u.r.obj) {
          gl_append_ptr(g_prob.objrels,(VOIDPTR)rip);
        } else {
          gl_append_ptr(g_prob.rels,(VOIDPTR)rip);
        }
      }
      break;
    case LREL_INST:
      rip = SIP(GetInterfacePtr(child));
      rip->u.lr.model = modindex;
      if (rip->u.lr.cond) {
        gl_append_ptr(g_prob.logcnds,(VOIDPTR)rip);
      }
      else {
        gl_append_ptr(g_prob.logrels,(VOIDPTR)rip);
      }
      break;
    case WHEN_INST:
      rip = SIP(GetInterfacePtr(child));
      rip->u.w.model = modindex;
      gl_append_ptr(g_prob.whens,(VOIDPTR)rip);
      break;
    case ARRAY_ENUM_INST:
    case ARRAY_INT_INST:
      if (ArrayIsRelation(child)|| ArrayIsLogRel(child)|| ArrayIsWhen(child)) {
	CollectArrayRelsAndWhens(child,modindex);
      }
      break;
    default:
      break;
    }
  }
}



/*
 * Checks the problem extrels list to see whether a cache has been
 * created for the given relation in the problem_t bridge.
 * If not will return NULL, else
 * will return the pointer to the cache. The nodestamp corresponding
 * to this relation is returned regardless.
 *
 */
static struct ExtRelCache
  *CheckIfCacheExists( struct Instance *relinst, int *nodestamp)
{
  struct ExtCallNode *ext;
  struct ExtRelCache *cache;
  CONST struct relation *gut;
  enum Expr_enum type;
  unsigned long len,c;

  gut = GetInstanceRelation(relinst,&type);
  assert(type==e_blackbox);
  ext = BlackBoxExtCall(gut);
  *nodestamp = ExternalCallNodeStamp(ext);
  len = gl_length(g_prob.extrels);

  for (c=1;c<=len;c++) {
    cache = (struct ExtRelCache *)gl_fetch(g_prob.extrels,c);
    if (cache->nodestamp == *nodestamp) {
      return cache;
    }
  }
  return NULL;
}

/*
 * analyze_CountRelation
 * Call only with good relation instances.
 * Count the instance into the required bin.
 */
static void analyze_CountRelation(struct Instance *inst)
{
  switch( RelationRelop(GetInstanceRelationOnly(inst)) ) {
  case e_maximize:
  case e_minimize:
    g_prob.no++;
    break;
  case e_less: case e_lesseq:
  case e_greater: case e_greatereq:
  case e_equal: case e_notequal:
    if ( RelationIsCond(GetInstanceRelationOnly(inst)) ) {
      g_prob.nc++;
    }
    else {
      g_prob.nr++;
      if ( GetInstanceRelationType(inst)==e_blackbox ) {
        g_prob.ne++;
      }
    }
    break;
  default:
    FPRINTF(stderr,"ERROR:  (analyze) analyze_CountRelation\n");
    FPRINTF(stderr,"        Unknown relation type.\n");
  }
}


/*
 * GetIntFromSymbol
 * Used for a WHEN statement. It intends to obtain an integer value
 * from a symbol value. Each symbol value is storaged in a symbol list.
 * It checks if the symval is already in the solver symbol list,
 * if it is, returns the integer corresponding to the position of symval
 * if it is not, appends the symval to the list and then returns the int
 * This is terrible inefficient as the number of symbols grows.
 * I am keeping it by now, are they going to be so many symbols in
 * whens anyway ?
 */

int GetIntFromSymbol(CONST char *symval,
		     struct gl_list_t *symbol_list)
{
  struct SymbolValues *entry,*dummy;
  unsigned long length;
  int len,c,value;
  int symbol_list_count;

  if (symbol_list != NULL) {
    len = gl_length(symbol_list);
    for (c=1; c<= len; c++) {
      dummy = (struct SymbolValues *)(gl_fetch(symbol_list,c));
      if (!strcmp(dummy->name,symval)) {
	return (dummy->value);
      }
    }
    symbol_list_count = len;
  } else {
    symbol_list_count = 0;
    symbol_list = gl_create(2L);
  }
  length = (unsigned long)strlen(symval);
  length++;
  symbol_list_count++;
  entry = (struct SymbolValues *)ascmalloc(sizeof(struct SymbolValues));
  entry->name = (char *)ascmalloc(length *sizeof(char));
  strcpy(entry->name,symval);
  entry->value = symbol_list_count;
  value = entry->value;
  gl_append_ptr(symbol_list,entry);
  return value;
}

void DestroySymbolValuesList(struct gl_list_t *symbol_list)
{
  struct SymbolValues *entry;
  int len,c;
    if (symbol_list != NULL) {
      len = gl_length(symbol_list);
      for (c=1; c<= len; c++) {
        entry = (struct SymbolValues *)(gl_fetch(symbol_list,c));
        ascfree((char *)entry->name); /* Do I need this ? */
        ascfree((char *)entry);
      }
      gl_destroy(symbol_list);
      symbol_list = NULL;
    }
}




/*
 * classify_instance : to be called by PushInterfacPtrs.
 *
 * This function classifies the given instance and appends into
 * the necessary list in the g_prob structure.
 * It also sets returns a pointer to the caller for insertion into
 * the interface_ptr slot.
 *
 * Note, we should be passing the ipbuf info via vp so we don't
 * need the nasty global variable and can be more thread safe.
 */
static void *classify_instance(struct Instance *inst,VOIDPTR vp)
{
  struct solver_ipdata *ip;
  CONST char *symval;
  switch( InstanceKind(inst) ) {
  case REAL_ATOM_INST:   		/* Variable or parameter or real */
    ip = analyze_getip();
    ip->i = inst;
    ip->u.v.incident = 0;
    ip->u.v.in_block = 0;
    ip->u.v.index = 0;
    ip->u.v.active = 0;
    if( solver_var(inst) ) {
      ip->u.v.solvervar = 1;
      ip->u.v.fixed = BooleanChildValue(inst,"fixed");
      gl_append_ptr(g_prob.vars,(POINTER)ip);
    } else {
      ip->u.v.fixed = 1;
      ip->u.v.solvervar=0;
      if (solver_par(inst)) {
        gl_append_ptr(g_prob.pars,(POINTER)ip);
      } else {
        gl_append_ptr(g_prob.unas,(POINTER)ip);
      }
    }
    return ip;
  case BOOLEAN_ATOM_INST:
    ip = analyze_getip();
    ip->i = inst;
    ip->u.dv.isconst = 0;
    ip->u.dv.distype = 0;
    ip->u.dv.incident = 0;
    ip->u.dv.index = 0;
    ip->u.dv.active = 0;
      ip->u.dv.value = GetBooleanAtomValue(inst);
    if( boolean_var(inst) ) {
      ip->u.dv.booleanvar = 1;
      ip->u.dv.fixed = BooleanChildValue(inst,"fixed");
    } else {
      ip->u.dv.fixed = 1;
      ip->u.dv.booleanvar=0;
    }
    if( LogRelationsCount(inst) || WhensCount(inst) ) {
      gl_append_ptr(g_prob.dvars,(POINTER)ip);
      if ( WhensCount(inst) ) {
        ip->u.dv.inwhen = 1;
      } else {
        ip->u.dv.inwhen = 0;
      }
    } else {
      gl_append_ptr(g_prob.dunas,(POINTER)ip);
    }
    return ip;
  case BOOLEAN_CONSTANT_INST:
    if (WhensCount(inst)) {
      ip = analyze_getip();
      ip->i = inst;
      ip->u.dv.isconst = 1;
      ip->u.dv.distype = 0;
      ip->u.dv.index = 0;
      ip->u.dv.incident = 0;
      ip->u.dv.booleanvar=0;
      ip->u.dv.fixed = 1;
      ip->u.dv.active = 0;
      ip->u.dv.value = GetBooleanAtomValue(inst);
      gl_append_ptr(g_prob.dvars,(POINTER)ip);
      return ip;
    } else {
      return NULL;
    }
  case INTEGER_ATOM_INST:
    if (WhensCount(inst)) {
      ip = analyze_getip();
      ip->i = inst;
      ip->u.dv.isconst = 0;
      ip->u.dv.distype = 1;
      ip->u.dv.index = 0;
      ip->u.dv.incident = 0;
      ip->u.dv.booleanvar=0;
      ip->u.dv.fixed = 0;
      ip->u.dv.active = 0;
      ip->u.dv.value = GetIntegerAtomValue(inst);
      gl_append_ptr(g_prob.dvars,(POINTER)ip);
      return ip;
    } else {
      return NULL;
    }
  case SYMBOL_ATOM_INST:
    if (WhensCount(inst)) {
      ip = analyze_getip();
      ip->i = inst;
      ip->u.dv.isconst = 0;
      ip->u.dv.distype = -1;
      ip->u.dv.index = 0;
      ip->u.dv.incident = 0;
      ip->u.dv.booleanvar=0;
      ip->u.dv.fixed = 0;
      ip->u.dv.active = 0;
      symval = GetSymbolAtomValue(inst);
      if (g_symbol_values_list == NULL) {
        g_symbol_values_list = gl_create(2L);
      }
      ip->u.dv.value = GetIntFromSymbol(symval,g_symbol_values_list);
      gl_append_ptr(g_prob.dvars,(POINTER)ip);
      return ip;
    } else {
      return NULL;
    }
  case INTEGER_CONSTANT_INST:
    if (WhensCount(inst)) {
      ip = analyze_getip();
      ip->i = inst;
      ip->u.dv.isconst = 1;
      ip->u.dv.distype = 1;
      ip->u.dv.index = 0;
      ip->u.dv.incident = 0;
      ip->u.dv.booleanvar=0;
      ip->u.dv.fixed = 1;
      ip->u.dv.active = 0;
      ip->u.dv.value = GetIntegerAtomValue(inst);
      gl_append_ptr(g_prob.dvars,(POINTER)ip);
      return ip;
    } else {
      return NULL;
    }
  case SYMBOL_CONSTANT_INST:
    if (WhensCount(inst)) {
      ip = analyze_getip();
      ip->i = inst;
      ip->u.dv.isconst = 1;
      ip->u.dv.distype = -1;
      ip->u.dv.index = 0;
      ip->u.dv.incident = 0;
      ip->u.dv.booleanvar=0;
      ip->u.dv.fixed = 1;
      ip->u.dv.active = 0;
      symval = GetSymbolAtomValue(inst);
      if (g_symbol_values_list == NULL) {
        g_symbol_values_list = gl_create(2L);
      }
      ip->u.dv.value = GetIntFromSymbol(symval,g_symbol_values_list);
      gl_append_ptr(g_prob.dvars,(POINTER)ip);
      return ip;
    } else {
      return NULL;
    }
  case REL_INST:               /* Relation (or conditional or objective) */
    ip = analyze_getip();
    ip->i = inst;
    ip->u.r.active = 1;
    switch(RelationRelop(GetInstanceRelationOnly(inst))) {
    case e_minimize:
      ip->u.r.obj = 1;
      break;
    case e_maximize:
      ip->u.r.obj = -1;
      break;
    default:
      ip->u.r.obj = 0;
      break;
    }
    if ( GetInstanceRelationType(inst)==e_blackbox ) {
      ip->u.r.ext = 1;
    } else {
      ip->u.r.ext = 0;
    }
    if ( RelationIsCond(GetInstanceRelationOnly(inst)) ) {
      ip->u.r.cond = 1;
    } else {
      ip->u.r.cond = 0;
    }
    if ( WhensCount(inst) ) {
      ip->u.r.inwhen = 1;
    } else {
      ip->u.r.inwhen = 0;
    }
    ip->u.r.included = BooleanChildValue(inst,"included");
    ip->u.r.model = 0;
    ip->u.r.index = 0;
    return ip;
    /* note we do NOT append it to lists here */
  case LREL_INST:        		/* LogRelation */
    ip = analyze_getip();
    ip->i = inst;
    ip->u.lr.active = 1;
    if ( LogRelIsCond(GetInstanceLogRel(inst)) ) {
      ip->u.lr.cond = 1;
    } else {
      ip->u.lr.cond = 0;
    }
    if ( WhensCount(inst) ) {
      ip->u.lr.inwhen = 1;
    } else {
      ip->u.lr.inwhen = 0;
    }
    ip->u.lr.included = BooleanChildValue(inst,"included");
    ip->u.lr.model = 0;
    ip->u.lr.index = 0;
    return ip;
    /* note we do NOT append it to lists here */
  case MODEL_INST:
    ip = analyze_getip();
    ip->i = inst;
    ip->u.m.index = 0;
    if ( WhensCount(inst) ) {
      ip->u.m.inwhen = 1;
    } else {
      ip->u.m.inwhen = 0;
    }
    gl_append_ptr(g_prob.models,(POINTER)ip);
    return ip;
  case WHEN_INST:
    ip = analyze_getip();
    ip->i = inst;
    ip->u.w.index = 0;
    if ( WhensCount(inst) ) {
      ip->u.w.inwhen = 1;
    } else {
      ip->u.w.inwhen = 0;
    }
    return ip;
    /* note we do NOT append it to lists here */
  default:
    return NULL;
  }
}

/*
 * make_problem
 *
 *  Makes variable/relations/when lists and objective function by heuristic.
 *  Now also makes a list of all relations that are objectives.
 *  This does not affect the semantics of the previous objective
 *  code.
 *  Do NOTHING in this function which can lead to floating point
 *  errors -- in particular because we must leave the interface ptrs
 *  in the state they were found.
 */
/*
 * This function sets g_bad_rel_in_list TRUE if it finds any unhappy
 * relations. All the rest of the code depends on ALL relations being
 * good, so don't disable the g_bad_rel_in_list feature.
 */
static void CountStuffInTree(struct Instance *inst)
{
  if (inst!=NULL) {
    switch (InstanceKind(inst)) {
    case REL_INST:
      if( GetInstanceRelationOnly(inst) == NULL ||
          GetInstanceRelationType(inst) == e_undefined) {
	/* guard against null relations, unfinished ones */
        FPRINTF(stderr,"ERROR: CountStuffInTree found bad relation\n");
        WriteInstanceName(stderr,inst,g_prob.root);
        FPRINTF(stderr,"\n");
        g_bad_rel_in_list = TRUE;
        return;
      }
      analyze_CountRelation(inst); /* increment according to classification */
      break;
    case REAL_ATOM_INST:
      if( solver_var(inst) && RelationsCount(inst)) {
        g_prob.nv++;
      } else {
        if (solver_par(inst) && RelationsCount(inst)) {
          /* never passes right now */
          g_prob.np++;
        } else {
          g_prob.nu++;
        }
      }
      /* The use of  RelationsCount is heuristic since vars may be
       * used in relations higher in the tree than the problem is rooted.
       */
      break;
    case BOOLEAN_ATOM_INST:
      if ( ( boolean_var(inst) && LogRelationsCount(inst) ) ||
	   WhensCount(inst) ) {
	g_prob.ndv++;
      }
      else {
	g_prob.nud++;
      }
      break;
    case BOOLEAN_CONSTANT_INST:
    case INTEGER_ATOM_INST:
    case INTEGER_CONSTANT_INST:
    case SYMBOL_ATOM_INST:
    case SYMBOL_CONSTANT_INST:
      if (WhensCount(inst)) {
        g_prob.ndv++;
      }
      break;
    case LREL_INST:
      if( GetInstanceLogRelOnly(inst) == NULL ) {
        FPRINTF(stderr,"ERROR: CountStuffInTree found bad logrel\n");
        WriteInstanceName(stderr,inst,g_prob.root);
        FPRINTF(stderr,"\n");
        g_bad_rel_in_list = TRUE;
        return;
      }
      if ( LogRelIsCond(GetInstanceLogRel(inst)) ) {
        g_prob.ncl++;
      }
      else {
        g_prob.nl++;
      }
      break;
    case WHEN_INST:
      g_prob.nw++;
      break;
    case MODEL_INST:
      g_prob.nm++;
      break;
    default:
      break;
    }
  }
}


/*
 * This takes the already derived counts,
 * allocates all the memory we need to allocate for master,
 * and builds the var/rel/MODEL/etc master lists.
 * filling in g_prob and ips as far as possible.
 * Returns 0 normally, or 1 if insufficient memory, 2 if nothing to do.
 * If 1, then the user should deallocate any partial results in
 * a separate cleanup function for g_prob.
 *
 * In particular, after this is done we have
 * vars with correct ip values for:
 * index;
 *  incident;
 *  in_block;
 *  fixed;	(as flag)
 *  solvervar;	(as flag)
 * relations and conditional relations with correct ip values for:
 *  index;
 *  model;
 *  obj;		(0 constraint, -1 maximize, +1 minimize)
 *  ext;
 *  inwhen;
 *  cond;
 *  included;	(as flag)
 * discrete vars with correct ip values for:
 *  index;
 *  incident;
 *  isconst;
 *  distype;
 *  fixed;	(as flag)
 *  booleanvar;	(as flag)
 *  inwhen;
 * logrelations and conditional logrelations with correct ip values for:
 *  index;
 *  model;
 *  included;	(as flag)
 *  inwhen;
 *  cond;
 * whens with correct ip values for:
 *  index;
 *  model;
 *  inwhen;
 * models with correct ip values for:
 *  index;
 *
 * Note that these are all indexed from 1, being stored in gllists.
 */
static int analyze_make_master_lists(void)
{
  long int c, len,v,vlen;
  CONST struct Instance *i;
  CONST struct relation *gut;
  CONST struct logrelation *lgut;
  struct solver_ipdata *vip;
  struct gl_list_t *scratch;
  size_t reqlen;
  int stat;
  reqlen = g_prob.nr + g_prob.no + g_prob.nc + g_prob.nl+ g_prob.ncl+  /*rels*/
       g_prob.nv + g_prob.ndv +
       g_prob.np + g_prob.nu +  g_prob.nud +                        /* atoms */
       g_prob.nw +                                                  /* when */
       g_prob.nm;				                    /* model */
  if (reqlen <1) return 2;

  /* CREATE memory for ips */
#ifdef NDEBUG
  stat = resize_ipbuf(reqlen,0);
#else
  stat = resize_ipbuf(reqlen,1);
#endif /* NDEBUG */
  if (stat) {
    FPRINTF(stderr,"Error: (analyze_make_master) Insufficient memory.\n");
    return 1;
  }

  g_prob.vars = gl_create(g_prob.nv);		/* variables */
  if (g_prob.vars == NULL) return 1;

  g_prob.dvars = gl_create(g_prob.ndv);		/* discrete variables */
  if (g_prob.dvars == NULL) return 1;

  g_prob.pars = gl_create(g_prob.np);		/* parameters */
  if (g_prob.pars == NULL) return 1;

  g_prob.unas = gl_create(g_prob.nu);		/* unattached */
  if (g_prob.unas == NULL) return 1;

  g_prob.dunas = gl_create(g_prob.nud);		/* discrete unattached */
  if (g_prob.dunas == NULL) return 1;

  g_prob.cnds = gl_create(g_prob.nc);		/* conditional relations */
  if (g_prob.cnds == NULL) return 1;

  g_prob.rels = gl_create(g_prob.nr);		/* relations */
  if (g_prob.rels == NULL) return 1;

  g_prob.logrels = gl_create(g_prob.nl);       	/* logical relations */
  if (g_prob.logrels == NULL) return 1;

  g_prob.logcnds = gl_create(g_prob.ncl);  	/* conditional logrelations */
  if (g_prob.logcnds == NULL) return 1;

  g_prob.extrels = gl_create(g_prob.ne);	/* extrelations */
  if (g_prob.extrels == NULL) return 1;

  g_prob.objrels = gl_create(g_prob.no);	/* objectives */
  if (g_prob.objrels == NULL) return 1;

  g_prob.whens = gl_create(g_prob.nw);		/* whens */
  if (g_prob.whens == NULL) return 1;

  g_prob.models = gl_create(g_prob.nm);		/* models */
  if (g_prob.models == NULL) return 1;

  /* decorate the instance tree with ips, collecting vars and models. */
  g_prob.oldips = PushInterfacePtrs(g_prob.root,classify_instance,
                                    g_reuse.ipcap,1,NULL);
  if (g_prob.oldips == NULL) {
    FPRINTF(stderr,"Error: (analyze) Insufficient memory.\n");
    return 1;
  }

  /* 
   * collect relations, objectives, logrels and whens recording the
   * MODEL number in each rel's ip and setting incidence.
   */
  len = gl_length(g_prob.models);
  for (c=1; c <= len; c++) {
    CollectRelsAndWhens(SIP(gl_fetch(g_prob.models,c)),c);
    SIP(gl_fetch(g_prob.models,c))->u.m.index = c;
  }
  if ((long)gl_length(g_prob.rels) != g_prob.nr ||
      (long)gl_length(g_prob.objrels) != g_prob.no ||
      (long)gl_length(g_prob.logrels) != g_prob.nl ||
      (long)gl_length(g_prob.whens) != g_prob.nw) {
    FPRINTF(ASCERR,"Warning: Mismatch in problem census and problem found\n");
    FPRINTF(ASCERR,"Rels: Counted %lu\t Found %ld\n",gl_length(g_prob.rels),
      g_prob.nr);
    FPRINTF(ASCERR,"Objs: Counted %lu\t Found %ld\n",gl_length(g_prob.objrels),
      g_prob.no);
    FPRINTF(ASCERR,"LogRels: Counted %lu\t Found %ld\n",
      gl_length(g_prob.logrels),g_prob.nl);
    FPRINTF(ASCERR,"Whens: Counted %lu\t Found %ld\n",gl_length(g_prob.whens),
      g_prob.nw);
  }
  /* 
   * relation list is now grouped by model, and the order will be
   * invariant with hardware and ascend invocation so long as
   * set FIRSTCHOICE holds in compilation.
   */
  /* mark vars in constraints incident  and index rels */
  len = gl_length(g_prob.rels);
  for (c=1; c <= len; c++) {
    SIP(gl_fetch(g_prob.rels,c))->u.r.index = c;
    gut = GetInstanceRelationOnly(SIP(gl_fetch(g_prob.rels,c))->i);
    vlen = NumberVariables(gut);
    for( v = 1; v <= vlen; v++ ) {
      i = RelationVariable(gut,v);
      SIP(GetInterfacePtr(i))->u.v.incident = 1;
    }
  }
  /* mark vars in objectives incident */
  len = gl_length(g_prob.objrels);
  for (c=1; c <= len; c++) {
    gut = GetInstanceRelationOnly(SIP(gl_fetch(g_prob.objrels,c))->i);
    vlen = NumberVariables(gut);
    for( v = 1; v <= vlen; v++ ) {
      i = RelationVariable(gut,v);
      SIP(GetInterfacePtr(i))->u.v.incident = 1;
    }
  }

  /* mark vars in conditional relations incident */
  len = gl_length(g_prob.cnds);
  for (c=1; c <= len; c++) {
    SIP(gl_fetch(g_prob.cnds,c))->u.r.index = c;
    gut = GetInstanceRelationOnly(SIP(gl_fetch(g_prob.cnds,c))->i);
    vlen = NumberVariables(gut);
    for( v = 1; v <= vlen; v++ ) {
      i = RelationVariable(gut,v);
      SIP(GetInterfacePtr(i))->u.v.incident = 1;
    }
  }

  /* mark dvars in logrels incident  and index logrels */
  len = gl_length(g_prob.logrels);
  for (c=1; c <= len; c++) {
    SIP(gl_fetch(g_prob.logrels,c))->u.lr.index = c;
    lgut = GetInstanceLogRelOnly(SIP(gl_fetch(g_prob.logrels,c))->i);
    vlen = NumberBoolVars(lgut);
    for( v = 1; v <= vlen; v++ ) {
      i = LogRelBoolVar(lgut,v);
      SIP(GetInterfacePtr(i))->u.dv.incident = 1;
    }
  }


  /* mark dvars in conditional logrels incident */
  len = gl_length(g_prob.logcnds);
  for (c=1; c <= len; c++) {
    SIP(gl_fetch(g_prob.logcnds,c))->u.lr.index = c;
    lgut = GetInstanceLogRelOnly(SIP(gl_fetch(g_prob.logcnds,c))->i);
    vlen = NumberBoolVars(lgut);
    for( v = 1; v <= vlen; v++ ) {
      i = LogRelBoolVar(lgut,v);
      SIP(GetInterfacePtr(i))->u.dv.incident = 1;
    }
  }

  /* mark index whens */
  len = gl_length(g_prob.whens);
  for (c=1; c <= len; c++) {
    SIP(gl_fetch(g_prob.whens,c))->u.w.index = c;
  }
  /* 
   * now we need to move all the nonincident vars off the var list
   * onto the unas list. It is easiest to make a new list and copy
   * the existing var list to either it if keep or unas if punt.
   * the same goes for parameters that aren't incident.
   * We don't do exactly the same for discrete variables because
   * many of them are not incident but they are used in when's
   */
  len = gl_length(g_prob.vars);
  g_prob.tmplist = gl_create(len);
  if (g_prob.tmplist == NULL) return 1;
  for (c=1; c <= len; c++) {
    /* dispose of var */
    vip = SIP(gl_fetch(g_prob.vars,c));
    if ( vip->u.v.incident == 1) {
      gl_append_ptr(g_prob.tmplist,vip);
    } else {
      gl_append_ptr(g_prob.unas,vip);
    }
  }
  gl_destroy(g_prob.vars);		/* punt the old list */
  g_prob.vars = g_prob.tmplist; 	/* make shortened list var list */
  g_prob.tmplist = NULL;

  len = gl_length(g_prob.pars);
  g_prob.tmplist = gl_create(len);
  if (g_prob.tmplist == NULL) return 1;
  for (c=1; c <= len; c++) {
    /* dispose of par */
    vip = SIP(gl_fetch(g_prob.pars,c));
    if ( vip->u.v.incident == 1) {
      gl_append_ptr(g_prob.tmplist,vip);
    } else {
      gl_append_ptr(g_prob.unas,vip);
    }
  }
  gl_destroy(g_prob.pars);		/* punt the old list */
  g_prob.pars = g_prob.tmplist; 	/* make shortened list par list */
  g_prob.tmplist = NULL;

  /*
   * discrete variables: take the incident dis vars in logrels first,
   * then append the dis vars which are used only in whens
   */
  g_prob.lrelinc = 0;
  len = gl_length(g_prob.dvars);
  if ( len > 0) {
    g_prob.tmplist = gl_create(len);
    scratch = gl_create(2L);
    if (g_prob.tmplist == NULL) return 1;
    for (c=1; c <= len; c++) {
      /* dispose of incident discrete vars */
      vip = SIP(gl_fetch(g_prob.dvars,c));
      if ( vip->u.dv.incident == 1) {
        gl_append_ptr(g_prob.tmplist,vip);
        g_prob.lrelinc++;                  /* Number of incident dis vars */
      } else {
        gl_append_ptr(scratch,vip);
      }
    }
    gl_destroy(g_prob.dvars);
    g_prob.dvars = g_prob.tmplist;
    g_prob.tmplist = NULL;
    /* append discrete non-boolean vars at the end of the list */
    len = gl_length(scratch);
    for (c=1; c <= len; c++) {
      vip = SIP(gl_fetch(scratch,c));
      gl_append_ptr(g_prob.dvars,vip);
    }
    gl_destroy(scratch);
    scratch = NULL;
  }
  return 0;
}


/*  
 *  This function cleans up an errant problem_t or a good one that we're
 *  done with. We should have set to null any pointers to memory we are
 *  keeping elsewhere before calling this.
 */
#define AFUN(ptr) if (ptr!=NULL) ascfree(ptr); (ptr) = NULL
#define ADUN(ptr) if (ptr!=NULL) gl_destroy(ptr); (ptr) = NULL
static void analyze_free_lists(void)
{
/* memory containing gl_lists of atomic structure pointers */
  if (g_prob.extrels != NULL) gl_free_and_destroy(g_prob.extrels);
/* gl_lists without memory items use ADUN */
  ADUN(g_prob.vars);
  ADUN(g_prob.dvars);
  ADUN(g_prob.pars);
  ADUN(g_prob.unas);
  ADUN(g_prob.dunas);
  ADUN(g_prob.rels);
  ADUN(g_prob.objrels);
  ADUN(g_prob.models);
  ADUN(g_prob.cnds);
  ADUN(g_prob.logrels);
  ADUN(g_prob.logcnds);
  ADUN(g_prob.whens);
/* blocks of memory use AFUN */
  AFUN(g_prob.blocks);
  AFUN(g_prob.reldata);
  AFUN(g_prob.objdata);
  AFUN(g_prob.condata);
  AFUN(g_prob.lrdata);
  AFUN(g_prob.logcondata);
  AFUN(g_prob.vardata);
  AFUN(g_prob.pardata);
  AFUN(g_prob.undata);
  AFUN(g_prob.disdata);
  AFUN(g_prob.undisdata);
  AFUN(g_prob.whendata);
  AFUN(g_prob.bnddata);
  AFUN(g_prob.relincidence);
  AFUN(g_prob.varincidence);
  AFUN(g_prob.logrelinciden);
  AFUN(g_prob.mastervl);
  AFUN(g_prob.masterdl);
  AFUN(g_prob.masterrl);
  AFUN(g_prob.masterol);
  AFUN(g_prob.mastercl);
  AFUN(g_prob.masterll);
  AFUN(g_prob.mastercll);
  AFUN(g_prob.masterpl);
  AFUN(g_prob.masterul);
  AFUN(g_prob.masterdul);
  AFUN(g_prob.masterwl);
  AFUN(g_prob.masterbl);
  AFUN(g_prob.solvervl);
  AFUN(g_prob.solverdl);
  AFUN(g_prob.solverrl);
  AFUN(g_prob.solverol);
  AFUN(g_prob.solvercl);
  AFUN(g_prob.solverll);
  AFUN(g_prob.solvercll);
  AFUN(g_prob.solverpl);
  AFUN(g_prob.solverul);
  AFUN(g_prob.solverdul);
  AFUN(g_prob.solverwl);
  AFUN(g_prob.solverbl);
  AFUN(g_prob.erlist);
}



/*
 *                           When Processing
 */

/*
 * This function receives as argument the list of values of each of the
 * CASES of a WHEN statement. The values in the list can be integer values,
 * symbol values, or boolean values. So, the goal of this function is to
 * obtain an equivalent list of ONLY integer values for such a list. In this
 * way, for example, the boolean value TRUE is equivalent to the integer
 * 1. The function GentIntFromSymbol is used to generate an integer value
 * which will be equivalent to a symbol value
 */
static
void ProcessValueList(struct Set *ValueList, int *value,
		      struct gl_list_t *symbol_list)
{
  CONST struct Expr *expr;
  struct Set *s;

  s = ValueList;
  if (ValueList!=NULL) {
    while (s != NULL) {
      expr = GetSingleExpr(s);
      switch(ExprType(expr)) {
        case e_boolean:
	  *value = ExprBValue(expr);
	  if (*value == 2) {   /* ANY */
	    *value = -2;
	  }
	  break;
        case e_int:
	  *value = ExprIValue(expr);
	  break;
        case e_symbol:
	  *value = GetIntFromSymbol(ExprSymValue(expr),symbol_list);
	  break;
        default:
	  break;
      }
      s = NextSet(s);
      value++;
    }
  } else {
    *value = -1;  /* OTHERWISE */
  }
}


/*
 * The next two functions are used because in the function
 * ProcessSolverWhens, the existence of MODELS or ARRAYs inside
 * a When Statement requires a recursive analysis.
 * See the explanation of such a function
 */
static
void ProcessArraysInWhens(struct Instance *cur_inst,
			  struct gl_list_t *rels,
			  struct gl_list_t *logrels,
			  struct gl_list_t *whens)
{
  struct rel_relation *rel;
  struct logrel_relation *lrel;
  struct w_when *w;
  struct Instance *child;
  struct solver_ipdata *ip;
  unsigned long c,nch;

  if (cur_inst==NULL) return;
  nch = NumberChildren(cur_inst);
  for (c=1;c<=nch;c++) {
    child = InstanceChild(cur_inst,c);
    if (child==NULL) continue;
    switch (InstanceKind(child)) {
    case REL_INST:
      ip = SIP(GetInterfacePtr(child));
      ip->u.r.active = 0;
      rel = ip->u.r.data;
      gl_append_ptr(rels,rel);
      break;
    case LREL_INST:
      ip = SIP(GetInterfacePtr(child));
      ip->u.lr.active = 0;
      lrel = ip->u.lr.data;
      gl_append_ptr(logrels,lrel);
      break;
    case WHEN_INST:
      ip = SIP(GetInterfacePtr(child));
      w = ip->u.w.data;
      gl_append_ptr(whens,w);
      when_set_inwhen(w,TRUE);
      break;
    case MODEL_INST:
      ProcessModelsInWhens(child,rels,logrels,whens);
      break;
    case ARRAY_ENUM_INST:
    case ARRAY_INT_INST:
      if (ArrayIsRelation(child) || ArrayIsWhen(child)
         || ArrayIsLogRel(child) || ArrayIsModel(child)) {
        ProcessArraysInWhens(child,rels,logrels,whens);
      }
      break;
    default:
      break;
    }
  }
}

static
void ProcessModelsInWhens(struct Instance *cur_inst, struct gl_list_t *rels,
			  struct gl_list_t *logrels, struct gl_list_t *whens)
{
  struct rel_relation *rel;
  struct logrel_relation *lrel;
  struct w_when *w;
  struct Instance *child;
  struct solver_ipdata *ip;
  unsigned long c,nch;

  if (cur_inst==NULL) return;
  nch = NumberChildren(cur_inst);
  for (c=1;c<=nch;c++) {
    child = InstanceChild(cur_inst,c);
    if (child==NULL) continue;
    switch (InstanceKind(child)) {
    case REL_INST:
      ip = SIP(GetInterfacePtr(child));
      ip->u.r.active = 0;
      rel = ip->u.r.data;
      gl_append_ptr(rels,rel);
      break;
    case LREL_INST:
      ip = SIP(GetInterfacePtr(child));
      ip->u.lr.active = 0;
      lrel = ip->u.lr.data;
      gl_append_ptr(logrels,lrel);
      break;
    case WHEN_INST:
      ip = SIP(GetInterfacePtr(child));
      w = ip->u.w.data;
      gl_append_ptr(whens,w);
      when_set_inwhen(w,TRUE);
      break;
    case MODEL_INST:
      ProcessModelsInWhens(child,rels,logrels,whens);
      break;
    case ARRAY_ENUM_INST:
    case ARRAY_INT_INST:
      if (ArrayIsRelation(child) || ArrayIsWhen(child)
         || ArrayIsLogRel(child) || ArrayIsModel(child)) {
        ProcessArraysInWhens(child,rels,logrels,whens);
      }
      break;
    default:
      break;
    }
  }
}


/*
 * The goal of this function is to fill in the list of cases and variables
 * of a w_when structure with the appropriate data. The information required
 * is provided by the corresponding when Instance generated in the compilation
 * time. So, what we do is:
 * 1) Obtain the list of variables and the list of cases from each
 *    WHEN intance.
 *    The list of variables is actually a list of pointers to instances
 * 2) From each CASE, obtain also the list of references. This list of
 * references contains a list of pointers to each relation,logrelation  and
 * model included inside the case.
 * 3) The pointers to the variables, relations, logrelations and models are
 * used to obtain the solver data associated with the compiled instances.
 * 4) Arrays and models are managed recursively with the two previous
 * functions.
 */
static
void ProcessSolverWhens(struct w_when *when,struct Instance *i)
{
  struct gl_list_t *scratch;
  struct gl_list_t *wvars;
  struct gl_list_t *ref;
  struct gl_list_t *rels;
  struct gl_list_t *logrels;
  struct gl_list_t *whens;
  struct gl_list_t *diswhens;
  struct Set *ValueList;
  struct Instance *cur_inst;
  struct Case *cur_case;
  struct solver_ipdata *ip;
  struct dis_discrete *dvar;
  struct rel_relation *rel;
  struct logrel_relation *lrel;
  struct w_when *w;
  struct when_case *cur_sol_case;
  int c,r,len,lref;
  int *value;

  scratch = GetInstanceWhenVars(i);
  len = gl_length(scratch);
  wvars = gl_create(len);
  when->dvars = wvars;
  for (c=1;c<=len;c++) {
    cur_inst = (struct Instance *)(gl_fetch(scratch,c));
    ip = SIP(GetInterfacePtr(cur_inst));
    dvar = ip->u.dv.data;
    if (dis_whens_list(dvar)==NULL) {
      diswhens = gl_create(2L);
      dis_set_whens_list(dvar,diswhens);
    } else {
      diswhens = dis_whens_list(dvar);
    }
    gl_append_ptr(diswhens,when);
    gl_append_ptr(when->dvars,dvar);
  }

  scratch = GetInstanceWhenCases(i);
  len = gl_length(scratch);
  when->cases = gl_create(len);
  for (c=1;c<=len;c++) {
    cur_sol_case = when_case_create(NULL);
    cur_case = (struct Case *)(gl_fetch(scratch,c));
    ValueList = GetCaseValues(cur_case);
    value = &(cur_sol_case->values[0]);
    if (g_symbol_values_list == NULL) {
      g_symbol_values_list = gl_create(2L);
    }
    ProcessValueList(ValueList,value,g_symbol_values_list);
    ref = GetCaseReferences(cur_case);
    lref = gl_length(ref);
    rels = gl_create(lref);  /* maybe allocating less than needed (models) */
    logrels = gl_create(lref);  /* maybe allocating less than needed */
    whens = gl_create(lref); /* maybe allocating more than needed */
    for (r=1;r<=lref;r++) {
      cur_inst = (struct Instance *)(gl_fetch(ref,r));
      switch(InstanceKind(cur_inst)){
      case REL_INST:
        ip = SIP(GetInterfacePtr(cur_inst));
	ip->u.r.active = 0;
        rel = ip->u.r.data;
        gl_append_ptr(rels,rel);
	break;
      case LREL_INST:
        ip = SIP(GetInterfacePtr(cur_inst));
	ip->u.lr.active = 0;
        lrel = ip->u.lr.data;
        gl_append_ptr(logrels,lrel);
	break;
      case WHEN_INST:
        ip = SIP(GetInterfacePtr(cur_inst));
        w = ip->u.w.data;
        gl_append_ptr(whens,w);
        when_set_inwhen(w,TRUE);
	break;
      case MODEL_INST:
	ProcessModelsInWhens(cur_inst,rels,logrels,whens);
        break;
      default:
	break;
      }
    }
    when_case_set_rels_list(cur_sol_case,rels);
    when_case_set_logrels_list(cur_sol_case,logrels);
    when_case_set_whens_list(cur_sol_case,whens);
    when_case_set_active(cur_sol_case,FALSE);
    gl_append_ptr(when->cases,cur_sol_case);
  }
}


/*
 * Next two functions are for needed in the reconfiguration of
 * conditional models
 */

/*
 * return 1 if the discrete var is a member of the when var list, else
 * return 0
 */
int dis_var_in_a_when(struct Instance *var, struct w_when *when)
{
  struct Instance *winst;

  winst = (struct Instance *)(when_instance(when));
  return VarFoundInWhen(var,winst);
}


/*
 * Determine if the conditional variable inst is part of the
 * variable list of some when in the when list.
 */
int varinst_found_in_whenlist(slv_system_t sys, struct Instance *inst)
{
  struct w_when **whenlist;
  struct w_when *when;
  int c;

  whenlist = slv_get_solvers_when_list(sys);
  for (c=0; whenlist[c]!=NULL; c++) {
    when = whenlist[c];
    if (dis_var_in_a_when(inst,when)) {
      return 1;
    }
  }
  return 0;
}



/*
 *                           Boundary Processing
 */

/*
 * Get the list or logrelation including a boundary (by means of a
 * SATISFIED term). This function look the structures in the compiler
 * side and make the same link in the solver side
 */
static
void GetListOfLogRels(struct bnd_boundary *bnd, struct Instance *inst)
{
  struct gl_list_t *logrels;
  unsigned long c,len;
  struct Instance *i;
  struct logrel_relation *lrel;
  struct solver_ipdata *lrip;

  len = LogRelationsCount(inst);

  if (len>0) {
    logrels = gl_create(len);
    for (c=1; c<=len; c++) {
      i = LogRelationsForInstance(inst,c);
      lrip = SIP(GetInterfacePtr(i));
      lrel = lrip->u.lr.data;
      gl_append_ptr(logrels,lrel);
    }
    bnd_set_logrels(bnd,logrels);
  }
  return;
}

/*
 * Get the tolerance used to define the satisfaction of a boundary
 * (Defined in the SATISFIED term)
 */
static
void GetTolerance(struct bnd_boundary *bnd)
{
  struct gl_list_t *logrels;
  unsigned long c,len;
  struct logrel_relation *lrel;
  struct Instance *i,*rel;
  double tolerance;

  rel = (struct Instance *)(rel_instance(bnd_rel(bnd_real_cond(bnd))));
  logrels = bnd_logrels(bnd);
  len = gl_length(logrels);
  for (c=1; c<=len; c++) {
    lrel = (struct logrel_relation *)(gl_fetch(logrels,c));
    i = (struct Instance *)(logrel_instance(lrel));
    if (FindTolInSatTermOfLogRel(i,rel,&tolerance )) {
      bnd_set_tolerance(bnd,tolerance);
      return;
    }
  }
}




/*
 * Here we roll the master lists and bridge data into relation/var/
 * logrelation/conditional/when etc. lists for the consumer.
 * Includes fixing up rel caches and initing flagbits as best we can.
 * includes setting master indices on rel/var/logrel/when etc.
 * returns 0 if ok, 1 if out of memory, 2 if the problem does not
 * contain at least one variable in one equation
 */
static int analyze_make_solvers_lists(void)
{
  CONST struct relation *gut;
  CONST struct logrelation *lgut;
  struct ExtRelCache *cache;
  struct Instance *i;
  struct Instance *i_r;
  struct solver_ipdata *rip, *vip;
  struct solver_ipdata *lrip, *dvip, *wip;
  struct var_variable **incidence = NULL;
  struct rel_relation **varincidence = NULL;
  struct dis_discrete **logincidence = NULL;
  struct var_variable *var;
  struct rel_relation *rel;
  struct dis_discrete *dvar;
  struct logrel_relation *lrel;
  struct bnd_boundary *bnd;
  struct w_when *when;
  int order,nnzold, nodestamp;
  int logorder,lognnzold;
  int c,len,v,vlen,r,found;
  uint32 flags;

  order = MAX(gl_length(g_prob.vars),gl_length(g_prob.rels));
  nnzold = g_prob.nnz = g_prob.nnztot = g_prob.nnzobj = g_prob.nnzcond = 0;
  g_prob.nrow = 0; /* number of included relations */
  for (c=1,len = gl_length(g_prob.rels); c <= len; c++) {
    rip = SIP(gl_fetch(g_prob.rels,c));
    gut = GetInstanceRelationOnly(rip->i);
    vlen = NumberVariables(gut);
    g_prob.nnztot += vlen;
    if (rip->u.r.included) {
      g_prob.nrow++;
      nnzold = g_prob.nnz;
      for( v = 1 ; v <= vlen; v++ ) {
        i = RelationVariable(gut,v);
        vip = SIP(GetInterfacePtr(i));
        if (!(vip->u.v.fixed)) {
          g_prob.nnz++;
        }
      }
      if (g_prob.nnz==nnzold) {
        FPRINTF(ASCWAR,"No free variables in included relation:\n");
        WriteInstanceName(ASCWAR,rip->i,g_prob.root);
      }
    }
  }
  for (c=1,len = gl_length(g_prob.objrels); c <= len; c++) {
    rip = SIP(gl_fetch(g_prob.objrels,c));
    gut = GetInstanceRelationOnly(rip->i);
    vlen = NumberVariables(gut);
    g_prob.nnzobj += vlen;
  }

  /* Conditional relations */
  for (c=1,len = gl_length(g_prob.cnds); c <= len; c++) {
    rip = SIP(gl_fetch(g_prob.cnds,c));
    gut = GetInstanceRelationOnly(rip->i);
    vlen = NumberVariables(gut);
    g_prob.nnzcond += vlen;
  }


  /* 
   * calculate the number of free and incident variables, ncol 
   * we put all the nonincident on the unas list, so just check fixed. 
   */
  for (c=1,len = gl_length(g_prob.vars); c <= len; c++) {
    vip = SIP(gl_fetch(g_prob.vars,c));
    if (!(vip->u.v.fixed)) g_prob.ncol++;
  }
  /* 
   * now, at last we have cols jacobian in the order we want the lists to
   * be handed to the solvers.
   */


  logorder = MAX((unsigned long)g_prob.lrelinc,gl_length(g_prob.logrels));
  lognnzold = g_prob.lognnz = g_prob.lrelincsize = 0;
  g_prob.lognrow = 0; /* number of included logrelations */
  for (c=1,len = gl_length(g_prob.logrels); c <= len; c++) {
    lrip = SIP(gl_fetch(g_prob.logrels,c));
    lgut = GetInstanceLogRelOnly(lrip->i);
    vlen = NumberBoolVars(lgut);
    g_prob.lrelincsize += vlen;
    if (lrip->u.lr.included) {
      g_prob.lognrow++;
      lognnzold = g_prob.lognnz;
      for( v = 1 ; v <= vlen; v++ ) {
        i = LogRelBoolVar(lgut,v);
        dvip = SIP(GetInterfacePtr(i));
        if (!(dvip->u.dv.fixed)) {
          g_prob.lognnz++;
        }
      }
      if (g_prob.lognnz==lognnzold) {
        FPRINTF(ASCWAR,
                "No free boolean variables in included logrelation:\n");
        WriteInstanceName(ASCWAR,rip->i,g_prob.root);
      }
    }
  }

  /* Conditional logrelations */
  for (c=1,len = gl_length(g_prob.logcnds); c <= len; c++) {
    lrip = SIP(gl_fetch(g_prob.logcnds,c));
    lgut = GetInstanceLogRelOnly(lrip->i);
    vlen = NumberBoolVars(lgut);
    g_prob.lrelincsize += vlen;
  }

  if (!(g_prob.nnztot+g_prob.nnzobj) && !(g_prob.lognnz)) {
    FPRINTF(ASCERR,
            "Problem should contain at least one variable and one relation\n");
    return 2;
  }
  /*
   * we want at least one variable in one obj or rel,
   * or at least one boolean variable in one logrel
   */


  /* calculate the number of free and incident boolean variables, logncol */
  for (c=1,len = g_prob.lrelinc; c <= len; c++) {
    dvip = SIP(gl_fetch(g_prob.dvars,c));
    if (!(dvip->u.dv.fixed)) g_prob.logncol++;
  }


  /* now malloc and build things, remember to punt the matrix soon */
  /* remember we must NEVER free these things individually. */

#define ALLOCVARDATA(p,n) (p) = (struct var_variable *)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct var_variable)) : NULL)
#define ALLOCRELDATA(p,n) (p) = (struct rel_relation *)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct rel_relation)) : NULL)
#define ALLOCDISVARDATA(p,n) (p) = (struct dis_discrete *)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct dis_discrete)) : NULL)
#define ALLOCLOGRELDATA(p,n) (p) = (struct logrel_relation *)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct logrel_relation)) : NULL)
#define ALLOCWHENDATA(p,n) (p) = (struct w_when *)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct w_when)) : NULL)
#define ALLOCBNDDATA(p,n) (p) = (struct bnd_boundary *)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct bnd_boundary)) : NULL)
  ALLOCVARDATA(g_prob.vardata,g_prob.nv);
  ALLOCVARDATA(g_prob.pardata,g_prob.np);
  ALLOCVARDATA(g_prob.undata,g_prob.nu);
  ALLOCDISVARDATA(g_prob.disdata,g_prob.ndv);
  ALLOCDISVARDATA(g_prob.undisdata,g_prob.nud);
  ALLOCRELDATA(g_prob.reldata,g_prob.nr);
  ALLOCRELDATA(g_prob.objdata,g_prob.no);
  ALLOCRELDATA(g_prob.condata,g_prob.nc);
  ALLOCLOGRELDATA(g_prob.lrdata,g_prob.nl);
  ALLOCLOGRELDATA(g_prob.logcondata,g_prob.ncl);
  ALLOCWHENDATA(g_prob.whendata,g_prob.nw);
  ALLOCBNDDATA(g_prob.bnddata,g_prob.nc+g_prob.ncl);

#define ALLOCVARLIST(p,n) (p) = (struct var_variable **)( \
  ((n)>0) ? ascmalloc((n)*sizeof(struct var_variable *)) : NULL)
#define ALLOCRELLIST(p,n) (p) = (struct rel_relation **)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct rel_relation *)) : NULL)
#define ALLOCDISVARLIST(p,n) (p) = (struct dis_discrete **)( \
  ((n)>0) ? ascmalloc((n)*sizeof(struct dis_discrete *)) : NULL)
#define ALLOCLOGRELLIST(p,n) (p) = (struct logrel_relation **)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct logrel_relation *)) : NULL)
#define ALLOCWHENLIST(p,n) (p) = (struct w_when **)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct w_when *)) : NULL)
#define ALLOCBNDLIST(p,n) (p) = (struct bnd_boundary **)( \
    ((n)>0) ? ascmalloc((n)*sizeof(struct bnd_boundary *)) : NULL)
  ALLOCVARLIST(g_prob.mastervl,g_prob.nv+1);
  ALLOCVARLIST(g_prob.masterpl,g_prob.np+1);
  ALLOCVARLIST(g_prob.masterul,g_prob.nu+1);
  ALLOCDISVARLIST(g_prob.masterdl,g_prob.ndv+1);
  ALLOCDISVARLIST(g_prob.masterdul,g_prob.nud+1);
  ALLOCRELLIST(g_prob.masterrl,g_prob.nr+1);
  ALLOCRELLIST(g_prob.masterol,g_prob.no+1);
  ALLOCRELLIST(g_prob.mastercl,g_prob.nc+1);
  ALLOCLOGRELLIST(g_prob.masterll,g_prob.nl+1);
  ALLOCLOGRELLIST(g_prob.mastercll,g_prob.ncl+1);
  ALLOCWHENLIST(g_prob.masterwl,g_prob.nw+1);
  ALLOCBNDLIST(g_prob.masterbl,g_prob.nc+g_prob.ncl+1);
  ALLOCVARLIST(g_prob.solvervl,g_prob.nv+1);
  ALLOCVARLIST(g_prob.solverpl,g_prob.np+1);
  ALLOCVARLIST(g_prob.solverul,g_prob.nu+1);
  ALLOCDISVARLIST(g_prob.solverdl,g_prob.ndv+1);
  ALLOCDISVARLIST(g_prob.solverdul,g_prob.nud+1);
  ALLOCRELLIST(g_prob.solverrl,g_prob.nr+1);
  ALLOCRELLIST(g_prob.solverol,g_prob.no+1);
  ALLOCRELLIST(g_prob.solvercl,g_prob.nc+1);
  ALLOCLOGRELLIST(g_prob.solverll,g_prob.nl+1);
  ALLOCLOGRELLIST(g_prob.solvercll,g_prob.ncl+1);
  ALLOCWHENLIST(g_prob.solverwl,g_prob.nw+1);
  ALLOCBNDLIST(g_prob.solverbl,g_prob.nc+g_prob.ncl+1);

  ALLOCVARLIST(g_prob.relincidence,g_prob.nnztot+g_prob.nnzobj +
	       g_prob.nnzcond);
  ALLOCDISVARLIST(g_prob.logrelinciden,g_prob.lrelincsize);
#define CHECKPTRSIZE(n,p) if ((n)>0 && (p)==NULL) return 1
#define CHECKPTR(p) if ((p)==NULL) return 1
  /* verify mem allocations. */
  CHECKPTRSIZE(g_prob.nv,g_prob.vardata);
  CHECKPTRSIZE(g_prob.np,g_prob.pardata);
  CHECKPTRSIZE(g_prob.nu,g_prob.undata);
  CHECKPTRSIZE(g_prob.ndv,g_prob.disdata);
  CHECKPTRSIZE(g_prob.nud,g_prob.undisdata);
  CHECKPTRSIZE(g_prob.nr,g_prob.reldata);
  CHECKPTRSIZE(g_prob.no,g_prob.objdata);
  CHECKPTRSIZE(g_prob.nc,g_prob.condata);
  CHECKPTRSIZE(g_prob.nl,g_prob.lrdata);
  CHECKPTRSIZE(g_prob.ncl,g_prob.logcondata);
  CHECKPTRSIZE(g_prob.nw,g_prob.whendata);
  CHECKPTRSIZE(g_prob.nc+g_prob.ncl,g_prob.bnddata);
  CHECKPTR(g_prob.mastervl);
  CHECKPTR(g_prob.masterpl);
  CHECKPTR(g_prob.masterul);
  CHECKPTR(g_prob.masterdl);
  CHECKPTR(g_prob.masterdul);
  CHECKPTR(g_prob.masterrl);
  CHECKPTR(g_prob.masterol);
  CHECKPTR(g_prob.mastercl);
  CHECKPTR(g_prob.masterll);
  CHECKPTR(g_prob.mastercll);
  CHECKPTR(g_prob.masterwl);
  CHECKPTR(g_prob.masterbl);
  CHECKPTR(g_prob.solvervl);
  CHECKPTR(g_prob.solverpl);
  CHECKPTR(g_prob.solverul);
  CHECKPTR(g_prob.solverdl);
  CHECKPTR(g_prob.solverdul);
  CHECKPTR(g_prob.solverrl);
  CHECKPTR(g_prob.solverol);
  CHECKPTR(g_prob.solvercl);
  CHECKPTR(g_prob.solverll);
  CHECKPTR(g_prob.solvercll);
  CHECKPTR(g_prob.solverwl);
  CHECKPTR(g_prob.solverbl);
  CHECKPTR(g_prob.relincidence);
  CHECKPTRSIZE(g_prob.lrelincsize,g_prob.logrelinciden);
  g_prob.relincsize = g_prob.nnztot+g_prob.nnzobj + g_prob.nnzcond;
  g_prob.relincinuse = 0;
  g_prob.lrelincinuse = 0;

  /* 
   * for c in varlist copy vardata. remember gllist # from 1 and data from 0 
   */
  /* 
   * for c in varlist set mastervl, solvervl pointer to point to data 
   */
  vlen = gl_length(g_prob.vars);
  for (v = 0; v < vlen; v++) {
    var = &(g_prob.vardata[v]);
    vip = SIP(gl_fetch(g_prob.vars,v+1));
    vip->u.v.data = var;
    var_set_instance(var,vip->i);
    var_set_mindex(var,v);
    var_set_sindex(var,v);
    flags = 0; /* all init to FALSE */
    /* turn on appropriate ones */
    if (vip->u.v.incident)  flags |= VAR_INCIDENT;
    if (vip->u.v.in_block)  flags |= VAR_INBLOCK;
    if (vip->u.v.fixed)     flags |= VAR_FIXED;
    if (vip->u.v.solvervar) flags |= VAR_SVAR;
    var_set_flags(var,flags);
    g_prob.mastervl[v] = var;
    g_prob.solvervl[v] = var;
  }
  g_prob.mastervl[vlen] = NULL; /* terminator */
  g_prob.solvervl[vlen] = NULL; /* terminator */
  /* 
   * for c in parlist copy pardata. remember gllist # from 1 and data from 0
   * for c in parlist set masterpl, solverpl pointer to point to data 
   */
  vlen = gl_length(g_prob.pars);
  for (v = 0; v < vlen; v++) {
    var = &(g_prob.pardata[v]);
    vip = SIP(gl_fetch(g_prob.pars,v+1));
    vip->u.v.data = var;
    var_set_instance(var,vip->i);
    var_set_mindex(var,v);
    var_set_sindex(var,v);
    flags = 0; /* all init to FALSE */
    /* turn on appropriate ones */
    if (vip->u.v.incident)  flags |= VAR_INCIDENT;
    if (vip->u.v.in_block)  flags |= VAR_INBLOCK;
    if (vip->u.v.fixed)     flags |= VAR_FIXED;
    var_set_flags(var,flags);
    g_prob.masterpl[v] = var;
    g_prob.solverpl[v] = var;
  }
  g_prob.masterpl[vlen] = NULL; /* terminator */
  g_prob.solverpl[vlen] = NULL; /* terminator */
  /* 
   * for c in unalist copy undata. remember gllist # from 1 and data from 0 
   * for c in unalist set masterul, solverul pointer to point to data 
   */
  vlen = gl_length(g_prob.unas);
  for (v = 0; v < vlen; v++) {
    var = &(g_prob.undata[v]);
    vip = SIP(gl_fetch(g_prob.unas,v+1));
    vip->u.v.data = var;
    var_set_instance(var,vip->i);
    var_set_mindex(var,v);
    var_set_sindex(var,v);
    flags = 0; /* all init to FALSE */
    /* turn on appropriate ones */
    if (vip->u.v.incident)  flags |= VAR_INCIDENT;
    if (vip->u.v.fixed)     flags |= VAR_FIXED;
    if (vip->u.v.solvervar) flags |= VAR_SVAR;
    /* others may be appropriate (PVAR) */
    var_set_flags(var,flags);
    g_prob.masterul[v] = var;
    g_prob.solverul[v] = var;
  }
  g_prob.masterul[vlen] = NULL; /* terminator */
  g_prob.solverul[vlen] = NULL; /* terminator */

  /* 
   * process the constraining relations
   * for v in rellist copy reldata and fix extrels. 
   */
  vlen = gl_length(g_prob.rels);
  for (v = 0; v < vlen; v++) {
    rel = &(g_prob.reldata[v]);
    rip = SIP(gl_fetch(g_prob.rels,v+1));
    rel = rel_create(rip->i,rel);
    rip->u.r.data = rel;
    rel_set_mindex(rel,v);
    rel_set_sindex(rel,v);
    rel_set_model(rel,rip->u.r.model-1);
    /* here set up the var list */
    gut = GetInstanceRelationOnly(rip->i);
    assert(gut!=NULL);
    len = NumberVariables(gut);
    if (len > 0) {
      incidence = get_incidence_space(len);
      for( c = 0; c < len; c++ ) {
        i = RelationVariable(gut,c+1);
        incidence[c] = SIP(GetInterfacePtr(i))->u.v.data;
      }
      rel_set_incidences(rel,len,incidence);
    } else {
      rel_set_incidences(rel,0,NULL);
    }
    if (rel_extnodeinfo(rel)) {
      cache = CheckIfCacheExists(rip->i,&nodestamp);
      if (cache) {
        rel_set_extcache(rel,cache);
      } else {
        cache = CreateCacheFromInstance(rip->i);
        gl_append_ptr(g_prob.extrels,(POINTER)cache);
        rel_set_extcache(rel,cache);
      }
    }
    flags = 0; /* all init to FALSE */
    /* TURN ON APPROPRIATE ONES */
    if (rip->u.r.included) flags |= (REL_INCLUDED | REL_INBLOCK);
    if (rip->u.r.ext) flags |= REL_BLACKBOX;
    if (rip->u.r.active) flags |= ( REL_ACTIVE | REL_INVARIANT);
    if (rip->u.r.inwhen) flags |= REL_INWHEN;
    if ( RelationRelop(GetInstanceRelationOnly(rip->i)) == e_equal ) {
      flags |= REL_EQUALITY;
    }
    rel_set_flags(rel,flags);
    /* for c in rellist set masterrl, solverrl pointer to point to data */
    g_prob.masterrl[v] = rel;
    g_prob.solverrl[v] = rel;
  }
  g_prob.masterrl[vlen] = NULL; /* terminator */
  g_prob.solverrl[vlen] = NULL; /* terminator */

  /* cobble together external rel list */
  len = gl_length(g_prob.extrels);
  g_prob.erlist = (struct ExtRelCache **)
    ascmalloc((1+len)*sizeof(struct ExtRelCache *));
  if (g_prob.erlist==NULL) return 1;
  for (c=1; c <= len; c++) {
    g_prob.erlist[c-1] = (struct ExtRelCache *)gl_fetch(g_prob.extrels,c);
  }
  g_prob.erlist[len] = NULL; /* terminator */

/*
 * for c in objlist copy objdata.
 * for c in objlist set masterrl, solverrl pointer to point to data.
 */
  /* 
   * process the objective relations
   * for v in objlist copy objdata 
   */
  vlen = gl_length(g_prob.objrels);
  found = 0;
  for (v = 0; v < vlen; v++) {
    rel = &(g_prob.objdata[v]);
    rip = SIP(gl_fetch(g_prob.objrels,v+1));
    rel = rel_create(rip->i,rel);
    rip->u.r.data = rel;
    rel_set_mindex(rel,v);
    rel_set_sindex(rel,v);
    rel_set_model(rel,rip->u.r.model-1);
    /* here set up the var list */
    gut = GetInstanceRelationOnly(rip->i);
    assert(gut!=NULL);
    len = NumberVariables(gut);
    if (len > 0) {
      incidence = get_incidence_space(len);
      for( c = 0; c < len; c++ ) {
        i = RelationVariable(gut,c+1);
        incidence[c] = SIP(GetInterfacePtr(i))->u.v.data;
      }
      rel_set_incidences(rel,len,incidence);
    } else {
      rel_set_incidences(rel,0,NULL);
    }
    /* black box objectives not supported. skip it */
    flags = 0; /* all init to FALSE */
    /* TURN ON APPROPRIATE ONES  */
    if (rip->u.r.included) flags |= (REL_INCLUDED | REL_INBLOCK | REL_ACTIVE);
    if (rip->u.r.obj < 0) flags |= REL_OBJNEGATE;
    rel_set_flags(rel,flags);
    /* for c in objrellist set masterol, solverol pointer to point to data */
    g_prob.masterol[v] = rel;
    g_prob.solverol[v] = rel;
    /* set objective to first included objective on list */
    if (!found && (rip->u.r.included)) {
      g_prob.obj = rel;
      found = 1;
    }
  }
  g_prob.masterol[vlen] = NULL; /* terminator */
  g_prob.solverol[vlen] = NULL; /* terminator */

  /* 
   * process the conditional relations
   * for v in cndlist copy conddata . 
   */
  vlen = gl_length(g_prob.cnds);
  for (v = 0; v < vlen; v++) {
    rel = &(g_prob.condata[v]);
    rip = SIP(gl_fetch(g_prob.cnds,v+1));
    rel = rel_create(rip->i,rel);
    rip->u.r.data = rel;
    rel_set_mindex(rel,v);
    rel_set_sindex(rel,v);
    rel_set_model(rel,rip->u.r.model-1);
    gut = GetInstanceRelationOnly(rip->i);
    assert(gut!=NULL);
    len = NumberVariables(gut);
    if (len > 0) {
      incidence = get_incidence_space(len);
      for( c = 0; c < len; c++ ) {
        i = RelationVariable(gut,c+1);
        incidence[c] = SIP(GetInterfacePtr(i))->u.v.data;
      }
      rel_set_incidences(rel,len,incidence);
    } else {
      rel_set_incidences(rel,0,NULL);
    }
    flags = 0; /* all init to FALSE */
    /* TURN ON APPROPRIATE ONES */
    if (rip->u.r.included) flags |= (REL_INCLUDED | REL_INBLOCK | REL_ACTIVE);
    if (rip->u.r.cond) flags |= REL_CONDITIONAL;
    if ( RelationRelop(GetInstanceRelationOnly(rip->i)) == e_equal ) {
      flags |= REL_EQUALITY;
    }
    rel_set_flags(rel,flags);
    /* for c in rellist set masterrl, solverrl pointer to point to data */
    g_prob.mastercl[v] = rel;
    g_prob.solvercl[v] = rel;
    /* initially in same order */
  }
  g_prob.mastercl[vlen] = NULL; /* terminator */
  g_prob.solvercl[vlen] = NULL; /* terminator */


  /* 
   * process discrete variables 
   * for c in dvarlist copy disdata. gllist # from 1 and data from 0
   * for c in dvarlist set masterdl, solverdl pointer to point to data 
   */
  vlen = gl_length(g_prob.dvars);
  for (v = 0; v < vlen; v++) {
    dvar = &(g_prob.disdata[v]);
    dvip = SIP(gl_fetch(g_prob.dvars,v+1));
    /*
    dvip->u.dv.data = dvar;
    dis_set_instance(dvar,dvip->i); */
    /* from here */
    dis_create(dvip->i,dvar);
    dvip->u.dv.data = dvar;
    /* to here */
    dis_set_mindex(dvar,v);
    dis_set_sindex(dvar,v);
    dis_set_value(dvar,dvip->u.dv.value);
    dis_set_previous_value(dvar,dvip->u.dv.value);
    switch (dvip->u.dv.distype) {
      case 0:
        dis_set_kind(dvar,e_dis_boolean_t);
        break;
      case 1:
        dis_set_kind(dvar,e_dis_integer_t);
        break;
      case -1:
	dis_set_kind(dvar,e_dis_symbol_t);
        break;
      default:
	break;
    }
    flags = 0; /* all init to FALSE */
    /* turn on appropriate ones */
    if (dvip->u.dv.isconst)  flags |= DIS_CONST;
    if (dvip->u.dv.incident)  flags |= DIS_INCIDENT;
    if (dvip->u.dv.inwhen)  flags |= DIS_INWHEN;
    if (dvip->u.dv.fixed)     flags |= DIS_FIXED;
    if (dvip->u.dv.booleanvar) flags |= DIS_BVAR;
    if (dis_kind(dvar) == e_dis_boolean_t) flags |= DIS_BOOLEAN;
    dis_set_flags(dvar,flags);
    g_prob.masterdl[v] = dvar;
    g_prob.solverdl[v] = dvar;
    /* initially master and solver look the same */
  }
  g_prob.masterdl[vlen] = NULL; /* terminator */
  g_prob.solverdl[vlen] = NULL; /* terminator */

  /* 
   * for c in dunalist copy undisdata. gllist # from 1 and data from 0
   * for c in dunalist set masterdul, solverdul pointer to point to data 
   */
  vlen = gl_length(g_prob.dunas);
  for (v = 0; v < vlen; v++) {
    dvar = &(g_prob.undisdata[v]);
    dvip = SIP(gl_fetch(g_prob.dunas,v+1));
    dis_create(dvip->i,dvar);
    dvip->u.dv.data = dvar;
    dis_set_mindex(dvar,v);
    dis_set_sindex(dvar,v);
    flags = 0; /* all init to FALSE */
    /* turn on appropriate ones */
    if (dvip->u.dv.fixed)     flags |= DIS_FIXED;
    if (dvip->u.dv.booleanvar) flags |= DIS_BVAR;
    dis_set_flags(dvar,flags);
    g_prob.masterdul[v] = dvar;
    g_prob.solverdul[v] = dvar;
  }
  g_prob.masterdul[vlen] = NULL; /* terminator */
  g_prob.solverdul[vlen] = NULL; /* terminator */


/*
 * for c in logrellist copy lrdata.
 * for c in logrellist set masterll, solverll pointer to point to data.
 */
  /* 
   * process the logical relations 
   * for v in logrellist copy lrdata 
   */
  vlen = gl_length(g_prob.logrels);
  for (v = 0; v < vlen; v++) {
    lrel = &(g_prob.lrdata[v]);
    lrip = SIP(gl_fetch(g_prob.logrels,v+1));
    lrel = logrel_create(lrip->i,lrel);
    lrip->u.lr.data = lrel;
    logrel_set_mindex(lrel,v);
    logrel_set_sindex(lrel,v);
    logrel_set_model(lrel,lrip->u.lr.model-1);
    /* here set up the dis var list */
    lgut = GetInstanceLogRelOnly(lrip->i);
    assert(lgut!=NULL);
    len = NumberBoolVars(lgut);
    if (len > 0) {
      logincidence = get_logincidence_space(len);
      for( c = 0; c < len; c++ ) {
        i = LogRelBoolVar(lgut,c+1);
        logincidence[c] = SIP(GetInterfacePtr(i))->u.dv.data;
      }
      logrel_set_incidences(lrel,len,logincidence);
    } else {
      logrel_set_incidences(lrel,0,NULL);
    }
    flags = 0; /* all init to FALSE */
    /* TURN ON APPROPRIATE ONES  */
    if (lrip->u.lr.included) flags |= LOGREL_INCLUDED;
    if (lrip->u.lr.active) flags |= LOGREL_ACTIVE;
    if (lrip->u.lr.inwhen) flags |= LOGREL_INWHEN;
    if ( LogRelRelop(GetInstanceLogRelOnly(lrip->i)) == e_boolean_eq ) {
      flags |= LOGREL_EQUALITY;
    }
    logrel_set_flags(lrel,flags);
    /* for c in logrellist set masterll, solverll pointer to point to data */
    g_prob.masterll[v] = lrel;
    g_prob.solverll[v] = lrel;
  }
  g_prob.masterll[vlen] = NULL; /* terminator */
  g_prob.solverll[vlen] = NULL; /* terminator */


  /* 
   * process the conditional logrelations 
   * for v in logcndlist copy logconddata 
   */
  vlen = gl_length(g_prob.logcnds);
  for (v = 0; v < vlen; v++) {
    lrel = &(g_prob.logcondata[v]);
    lrip = SIP(gl_fetch(g_prob.logcnds,v+1));
    lrel = logrel_create(lrip->i,lrel);
    lrip->u.lr.data = lrel;
    logrel_set_mindex(lrel,v);
    logrel_set_sindex(lrel,v);
    logrel_set_model(lrel,lrip->u.lr.model-1);
    lgut = GetInstanceLogRelOnly(lrip->i);
    assert(lgut!=NULL);
    len = NumberBoolVars(lgut);
    if (len > 0) {
      logincidence = get_logincidence_space(len);
      for( c = 0; c < len; c++ ) {
        i = LogRelBoolVar(lgut,c+1);
        logincidence[c] = SIP(GetInterfacePtr(i))->u.dv.data;
      }
      logrel_set_incidences(lrel,len,logincidence);
    } else {
      logrel_set_incidences(lrel,0,NULL);
    }
    flags = 0; /* all init to FALSE */
    /* TURN ON APPROPRIATE ONES */
    if (lrip->u.lr.included) flags |= (LOGREL_INCLUDED | LOGREL_ACTIVE);
    if (lrip->u.lr.cond) flags |= LOGREL_CONDITIONAL;
    if ( LogRelRelop(GetInstanceLogRelOnly(lrip->i)) == e_boolean_eq) {
      flags |= LOGREL_EQUALITY;
    }
    logrel_set_flags(lrel,flags);
    /* for c in lrellist set masterll, solverll pointer to point to data */
    g_prob.mastercll[v] = lrel;
    g_prob.solvercll[v] = lrel;
    /* initially in same order */
  }
  g_prob.mastercll[vlen] = NULL; /* terminator */
  g_prob.solvercll[vlen] = NULL; /* terminator */


  /* 
   * process the boundaries
   * for v in cndlist and logcndlist, copy bnddata. 
   */
  vlen = gl_length(g_prob.cnds);
  len = gl_length(g_prob.logcnds);
  /* real conditions  */
  for (v = 0; v < vlen; v++) {
    bnd = &(g_prob.bnddata[v]);
    bnd = bnd_create(bnd);
    bnd_set_kind(bnd,e_bnd_rel);
    rip = SIP(gl_fetch(g_prob.cnds,v+1));
    bnd_real_cond(bnd) = bnd_rel(rip->u.r.data);
    bnd_set_mindex(bnd,v);
    bnd_set_sindex(bnd,v);
    bnd_set_model(bnd,rip->u.r.model-1);
    flags = 0; /* all init to FALSE */
    flags |= BND_REAL;
    bnd_set_flags(bnd,flags);
    /* for c in lrellist set masterbl, solverbl pointer to point to data */
    g_prob.masterbl[v] = bnd;
    g_prob.solverbl[v] = bnd;
  }
  /* logical conditions */
  for (v = vlen; v <vlen+len; v++) {
    bnd = &(g_prob.bnddata[v]);
    bnd = bnd_create(bnd);
    bnd_set_kind(bnd,e_bnd_logrel);
    lrip = SIP(gl_fetch(g_prob.logcnds,v-vlen+1));
    bnd_log_cond(bnd) = bnd_logrel(lrip->u.lr.data);
    bnd_set_mindex(bnd,v);
    bnd_set_sindex(bnd,v);
    bnd_set_model(bnd,lrip->u.lr.model-1);
    flags = 0; /* all init to FALSE */
    bnd_set_flags(bnd,flags);
    /* for c in lrellist set masterbl, solverbl pointer to point to data */
    g_prob.masterbl[v] = bnd;
    g_prob.solverbl[v] = bnd;
  }
  g_prob.masterbl[vlen+len] = NULL; /* terminator */
  g_prob.solverbl[vlen+len] = NULL; /* terminator */

  /*
   * Finding list of logical relations using the condition, and the
   * tolerance (only for the case of real condition ). Defining some
   * flags
   */

  for (v = 0; v < vlen; v++) {
    bnd = g_prob.masterbl[v];
    rel = bnd_rel(bnd_real_cond(bnd));
    flags = bnd_flags(bnd);
    if(rel_equality(rel)) {
      flags |= BND_EQUALITY;
    }
    i =  (struct Instance *)rel_instance(rel);
    GetListOfLogRels(bnd,i);
    if(bnd_logrels(bnd)!= NULL) {
      flags |= BND_IN_LOGREL;
      GetTolerance(bnd);
    }
    bnd_set_flags(bnd,flags);
  }
  for (v = vlen; v < vlen+len; v++) {
    bnd = g_prob.masterbl[v];
    lrel = bnd_logrel(bnd_log_cond(bnd));
    flags = bnd_flags(bnd);
    if(logrel_equality(lrel)) {
      flags |= BND_EQUALITY;
    }
    i =  (struct Instance *)logrel_instance(lrel);
    GetListOfLogRels(bnd,i);
    if(bnd_logrels(bnd)!= NULL) {
      flags |= BND_IN_LOGREL;
    }
    bnd_set_flags(bnd,flags);
  }

/*
 * for c in whenlist copy whendata.
 * for c in whenllist set masterwl, solverwl pointer to point to data.
 */
  /* process whens */

  vlen = gl_length(g_prob.whens);
  for (v = 0; v < vlen; v++) {
    when = &(g_prob.whendata[v]);
    wip = SIP(gl_fetch(g_prob.whens,v+1));
    when = when_create(wip->i,when);
    wip->u.w.data = when;
    when_set_mindex(when,v);
    when_set_sindex(when,v);
    when_set_model(when,wip->u.w.model-1);
    g_prob.masterwl[v] = when;
    g_prob.solverwl[v] = when;
    flags = 0;
    if (wip->u.w.inwhen) flags |= WHEN_INWHEN;
    when_set_flags(when,flags);
  }
  g_prob.masterwl[vlen] = NULL; /* terminator */
  g_prob.solverwl[vlen] = NULL; /* terminator */

  /* 
   * Get data from the when instance to fill the
   * list in the w_when instance 
   */

  for (v = 0; v < vlen; v++) {
    when = g_prob.masterwl[v];
    i = (struct Instance *)(when_instance(when));
    ProcessSolverWhens(when,i);
  }

  /* configure the problem */

  if (vlen > 0) { /* we have whens */

    configure_conditional_problem(vlen,g_prob.masterwl,
                                  g_prob.solverrl,g_prob.solverll);

  } else {

    /*
     * All variables in active relations are set as active.
     * This is necessary because the existence of some variables
     * in conditional relations which should not be active.
     *
     * Before we were doing:
     *
     *  for (v = 0;g_prob.solvervl[v]!=NULL ; v++) {
     *    var = g_prob.solvervl[v];
     *    var_set_active(var,TRUE);
     *  }
     *
     *  for (v = 0;g_prob.solverdl[v]!=NULL ; v++) {
     *    dvar = g_prob.solverdl[v];
     *    dis_set_active(dvar,TRUE);
     *  }
     *
     * which do not considerate such situation
     */

     set_active_vars_in_active_rels(g_prob.solverrl);
     set_active_disvars_in_active_logrels(g_prob.solverll);

    /*
     * All the unattached are set active to keep an accurate
     * counting in the solver side.
     */

    for (v = 0;g_prob.solverul[v]!=NULL ; v++) {
      var = g_prob.solverul[v];
      var_set_active(var,TRUE);
    }

    for (v = 0;g_prob.solverdul[v]!=NULL ; v++) {
      dvar = g_prob.solverdul[v];
      dis_set_active(dvar,TRUE);
    }
  }

  /*
   * Code to make the variables aware of the relation they are
   * incident in. KHT
   */
  vlen = gl_length(g_prob.vars);
  for (v = 0; v < vlen; v++) {
    var = &(g_prob.vardata[v]);
    i = var_instance(var);
    len = RelationsCount(i);
    g_prob.varincsize += len;
  }

  ALLOCRELLIST(g_prob.varincidence,g_prob.varincsize);

  vlen = gl_length(g_prob.vars);
  for (v = 0; v < vlen; v++) {
    var = &(g_prob.vardata[v]);
    i = var_instance(var);
    len = RelationsCount(i);
    r = 0;
    if (len > 0) {
      varincidence = get_var_incidence_space(len);
      for( c = 1; c <= len; c++ ) {
        i_r = RelationsForAtom(i,c);
	if( i_r == rel_instance(GetInterfacePtr(i_r)) ) {
	  varincidence[r] = SIP(GetInterfacePtr(i_r))->u.r.data;
	  r++;
	}
      }
    }
    if (r > 0) {
      var_set_incidences(var,r,varincidence);
    } else {
      var_set_incidences(var,0,NULL);
    }
  }
  return 0;
}



/*
 * hand off all the null terminated arrays to slv_system_t.
 * Also makes sure all solver_var have been assigned at least once,
 * since 0 is a really stupid starting value.
 */
static int analyze_configure_system(slv_system_t sys)
{
  slv_set_var_buf(sys,g_prob.vardata);
  g_prob.vardata = NULL;
  slv_set_par_buf(sys,g_prob.pardata);
  g_prob.pardata = NULL;
  slv_set_dvar_buf(sys,g_prob.disdata,gl_length(g_prob.dvars));
  g_prob.disdata = NULL;
  slv_set_rel_buf(sys,g_prob.reldata);
  g_prob.reldata = NULL;
  slv_set_condrel_buf(sys,g_prob.condata);
  g_prob.condata = NULL;
  slv_set_obj_buf(sys,g_prob.objdata);
  g_prob.objdata = NULL;
  slv_set_logrel_buf(sys,g_prob.lrdata);
  g_prob.lrdata = NULL;
  slv_set_condlogrel_buf(sys,g_prob.logcondata);
  g_prob.logcondata = NULL;
  slv_set_when_buf(sys,g_prob.whendata,gl_length(g_prob.whens));
  g_prob.whendata = NULL;
  slv_set_bnd_buf(sys,g_prob.bnddata,
                            gl_length(g_prob.cnds)+gl_length(g_prob.logcnds));
  g_prob.bnddata = NULL;
  slv_set_unattached_buf(sys,g_prob.undata);
  g_prob.undata = NULL;
  slv_set_disunatt_buf(sys,g_prob.undisdata);
  g_prob.undisdata = NULL;
  slv_set_incidence(sys,g_prob.relincidence,g_prob.relincsize);
  g_prob.relincidence = NULL;
  slv_set_var_incidence(sys,g_prob.varincidence,g_prob.varincsize);
  g_prob.varincidence = NULL;
  slv_set_logincidence(sys,g_prob.logrelinciden,g_prob.lrelincsize);
  g_prob.logrelinciden = NULL;
  slv_set_symbol_list(sys,g_symbol_values_list);
  g_symbol_values_list = NULL;
  slv_set_master_var_list(sys,g_prob.mastervl,gl_length(g_prob.vars));
  g_prob.mastervl = NULL;
  slv_set_master_par_list(sys,g_prob.masterpl,gl_length(g_prob.pars));
  g_prob.masterpl = NULL;
  slv_set_master_dvar_list(sys,g_prob.masterdl,gl_length(g_prob.dvars));
  g_prob.masterdl = NULL;
  slv_set_master_rel_list(sys,g_prob.masterrl,gl_length(g_prob.rels));
  g_prob.masterrl = NULL;
  slv_set_master_condrel_list(sys,g_prob.mastercl,gl_length(g_prob.cnds));
  g_prob.mastercl = NULL;
  slv_set_master_obj_list(sys,g_prob.masterol,gl_length(g_prob.objrels));
  g_prob.masterol = NULL;
  slv_set_master_logrel_list(sys,g_prob.masterll,gl_length(g_prob.logrels));
  g_prob.masterll = NULL;
  slv_set_master_condlogrel_list(sys,g_prob.mastercll,
                                                 gl_length(g_prob.logcnds));
  g_prob.mastercll = NULL;
  slv_set_master_when_list(sys,g_prob.masterwl,gl_length(g_prob.whens));
  g_prob.masterwl = NULL;
  slv_set_master_bnd_list(sys,g_prob.masterbl,
                          gl_length(g_prob.cnds)+gl_length(g_prob.logcnds));
  g_prob.masterbl = NULL;
  slv_set_master_unattached_list(sys,g_prob.masterul,gl_length(g_prob.unas));
  g_prob.masterul = NULL;
  slv_set_master_disunatt_list(sys,g_prob.masterdul,gl_length(g_prob.dunas));
  g_prob.masterdul = NULL;

  slv_set_solvers_var_list(sys,g_prob.solvervl,gl_length(g_prob.vars));
  g_prob.solvervl = NULL;
  slv_set_solvers_par_list(sys,g_prob.solverpl,gl_length(g_prob.pars));
  g_prob.solverpl = NULL;
  slv_set_solvers_dvar_list(sys,g_prob.solverdl,gl_length(g_prob.dvars));
  g_prob.solverdl = NULL;
  slv_set_solvers_rel_list(sys,g_prob.solverrl,gl_length(g_prob.rels));
  g_prob.solverrl = NULL;
  slv_set_solvers_condrel_list(sys,g_prob.solvercl,gl_length(g_prob.cnds));
  g_prob.solvercl = NULL;
  slv_set_solvers_obj_list(sys,g_prob.solverol,gl_length(g_prob.objrels));
  g_prob.solverol = NULL;
  slv_set_solvers_logrel_list(sys,g_prob.solverll,gl_length(g_prob.logrels));
  g_prob.solverll = NULL;
  slv_set_solvers_condlogrel_list(sys,g_prob.solvercll,
                                                  gl_length(g_prob.logcnds));
  g_prob.solvercll = NULL;
  slv_set_solvers_when_list(sys,g_prob.solverwl,gl_length(g_prob.whens));
  g_prob.solverwl = NULL;
  slv_set_solvers_bnd_list(sys,g_prob.solverbl,
                          gl_length(g_prob.cnds)+gl_length(g_prob.logcnds));
  g_prob.solverbl = NULL;
  slv_set_solvers_unattached_list(sys,g_prob.solverul,gl_length(g_prob.unas));
  g_prob.solverul = NULL;
  slv_set_solvers_disunatt_list(sys,g_prob.solverdul,gl_length(g_prob.dunas));
  g_prob.solverdul = NULL;

  slv_set_obj_relation(sys,g_prob.obj);
  g_prob.obj = NULL;

  slv_set_extrel_list(sys,g_prob.erlist,gl_length(g_prob.extrels));
  g_prob.erlist = NULL;

  slv_set_num_models(sys,g_prob.nm);

  PopInterfacePtrs(g_prob.oldips,NULL,NULL);
  g_prob.oldips = NULL;
  return 0;
}

/*
 * fills in a slv_system_t via its object interface from the
 * ascend side of the world, establishing any protocols needed to
 * communicate with the instance tree.
 * Return is 0 if everything ok, nonzero OTHERWISE.
 * 1 = memory
 * 2 = bad instance
 */
int analyze_make_problem(slv_system_t sys, struct Instance *inst)
{
  int stat;
  g_bad_rel_in_list = FALSE;
  InitTreeCounts(inst);
  /* take the census */
  VisitInstanceTree(inst,CountStuffInTree,TRUE,FALSE);
  if (g_bad_rel_in_list) {
    g_prob.root = NULL;
    return 2;
  }

  /* decorate instances with temporary ips, collect them and etc */
  stat = analyze_make_master_lists();
  if (stat == 2) {
    analyze_free_lists();
    FPRINTF(stderr,"Error: (%s) Nothing to make a problem from.\n",__FILE__);
    return 2;
  }
  if (stat == 1) {
    analyze_free_lists();
    FPRINTF(stderr,"Error: (%s) Insufficient master memory.\n",__FILE__);
    return 1;
  }
  /* rearrange all the stuff we found and index things */
  stat = analyze_make_solvers_lists();
  if (stat == 2) {
    analyze_free_lists();
    FPRINTF(stderr,"Error: (%s) Nothing to make a problem from.\n",__FILE__);
    return 2;
  }
  if (stat == 1) {
    analyze_free_lists();
    FPRINTF(stderr,"Error: (%s) Insufficient solver memory.\n",__FILE__);
    return 1;
  }

  /* tell the slv_system_t about it, and undecorate ips from instances */
  analyze_configure_system(sys);

  /* reordering things ala kirk is now a clients job */
  return 0;
}

extern void analyze_free_reused_mem(void)
{
  resize_ipbuf((size_t)0,0);
  analyze_free_lists();
}
