/*
 *  SLV: Ascend Numeric Solver
 *  by Karl Michael Westerberg
 *  Created: 2/6/90
 *  Version: $Revision: 1.19 $
 *  Version control file: $RCSfile: slv2.c,v $
 *  Date last modified: $Date: 97/07/18 12:16:00 $
 *  Last modified by: $Author: mthomas $
 *
 *  This file is part of the SLV solver.
 *
 *  Copyright (C) 1990 Karl Michael Westerberg
 *  Copyright (C) 1994 Joseph Zaher, Benjamin Andrew Allan
 *
 *  The SLV solver is free software; you can redistribute
 *  it and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  The SLV solver is distributed in hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge, MA 02139 USA.  Check the file named
 *  COPYING.  COPYING is found in ../compiler.
 *
 */

#include "utilities/ascConfig.h"
#include "utilities/ascMalloc.h"
#include "general/time.h"
#include "utilities/set.h"
#include "utilities/mem.h"
#include "compiler/compiler.h"
#include "compiler/fractions.h"
#include "general/list.h"
#include "compiler/extfunc.h"
#include "compiler/dimen.h"
#include "compiler/types.h"
#include "compiler/find.h"
#include "compiler/relation_type.h"
#include "compiler/relation.h"  /* relation enum */
#include "solver/mtx.h"
#include "solver/linsol.h"
#include "solver/linsolqr.h"
#include "solver/slv_types.h"
#include "solver/var.h"
#include "solver/rel.h"
#include "solver/discrete.h"
#include "solver/conditional.h"
#include "solver/logrel.h"
#include "solver/bnd.h"
#include "solver/calc.h"
#include "solver/relman.h"
#include "solver/slv_common.h"
#include "solver/slv_client.h"
#include "solver/slv2.h"

#if !defined(STATIC_OPTSQP) && !defined(DYNAMIC_OPTSQP)
int slv2_register(SlvFunctionsT *f)
{
  (void)f;  /* stop gcc whine about unused parameter */

  FPRINTF(stderr,"Opt not compiled in this ASCEND IV.\n");
  return 1;
}
#else /* either STATIC_OPTSQP or DYNAMIC_OPTSQP is defined */
#ifdef DYNAMIC_OPTSQP
/* do dynamic loading stuff.   yeah, right */
#else /* following is used if STATIC_OPTSQP is defined */

/*********************************************************************

 OUTPUT.MORE_IMPORTANT:
 OUTPUT.LESS_IMPORTANT: These will be used by the C interface and the
   jacobian and function calls. The control of MINOS noise is done via
   sub_parameter.
 TOLERANCE.PIVOT:        minos has no semantic equivalent.
 TOLERANCE.SINGULAR:     as minos PIVOT TOLERANCE
 TOLERANCE.FEASIBLE:     as minos ROW TOLERANCE & FEASIBILITY TOLERANCE
 TOLERANCE.STATIONARY:   as minos OPTIMALITY TOLERANCE
 TOLERANCE.TERMINATION:  no semantic equivalent.
 TIME_LIMIT:             use in C same as for slv0
 ITERATION_LIMIT:        as minos MAJOR ITERATIONS
 PARTITION:              minos has no semantic equivalent.
 IGNORE_BOUNDS:          minos disallows this entirely.
 RHO:                    as minos PENALTY PARAMETER

******************  SUBPARAMETERS FOR LB_OPT  ****************************

 Subparameter implemented:	(value/meaning)

 sp.ia[SP2_OPTION]		1 -> linesearch only
				2 -> trustregion only
 				3 -> linesearch and trustregion
   				     (NOTE: Only does 1 now)
 sp.ia[SP2_ISCALE]		0 -> doesn't scale objective function
				1 -> scales objective function
 sp.ia[SP2_ICHOOSE]		0 -> MA28 basis
				1 -> first m of independent
				2 -> last m of independent
 				3 -> get independent vars from d.dat
				4 -> get dependent vars from d.dat
				5 -> undecomposed
 sp.ia[SP2_IMULT]		0 -> with multipliers
				1 -> without multipliers
 sp.ia[SP2_ISAFE]		0 -> Use theory
				1 -> Use heuristic updating of penalty parameter
				     (NOTE: Only does 0 now)
 sp.ia[SP2_ICORR]		0 -> No correction term
				1 -> Finite difference using multipliers
				2 -> Finite difference without multipliers
				3 -> Broyden update
				4 -> Limited memory Broyden
				5 -> Interactive correction
 sp.ia[SP2_KPRINT]		Print output -> default = 1
 sp.ia[SP2_IIEXACT]		0 -> No exact penalty function
				1 -> Exact penalty function
				2 -> Switches between exact & augmented lagrangian
 sp.ia[SP2_IDEBUG]		0 -> no debug info printed
				1 -> debug info printed

 sp.ra[SP2_EPS]          	Convergence tolerance on K-T error
				  default = 10e-7
 sp.ra[SP2_VV]			Trust radius multipliers
				  default = 0.1

***********************************************************************/


/**********************************\
* NOUNDERBARS --> FORTRAN compiler *
* naming convention for subroutine *
* is wierd.                        *
* CRAY is treated as special case  *
\**********************************/
#ifdef APOLLO
#define NOUNDERBARS TRUE
#endif
#ifdef _HPUX_SOURCE
#define NOUNDERBARS TRUE
#endif

#ifdef NOUNDERBARS
#define FUNGRAD_SUB_NAME fungrad
#define FUNOBJ_SUB_NAME funobj
#define LB_OPT lb_opt
#else
#define FUNGRAD_SUB_NAME fungrad_
#define FUNOBJ_SUB_NAME funobj_
#define LB_OPT lb_opt_
#endif

#ifdef CRAY
#define FUNGRAD_SUB_NAME FUNGRAD
#define FUNOBJ_SUB_NAME FUNOBJ
#define LB_OPT LB_OPT
#endif

/* ALL LB_OPT ints below are expected to be 32 bits (INTEGER*4) */
/* len on char * are length expected not counting null         */

extern  LB_OPT (int *,		/* n */
		int *,		/* m */
		int *,		/* md */
		double *,	/* x */
		double *,	/* f */
		double *,	/* g */
		double *,	/* c */
		double *,	/* a */
		int *,		/* nz */
		int *,		/* nzd */
		int *,		/* avar */
		int *,		/* acon */
		int *,		/* ibnds */
		double *,	/* bnds */
		double *,	/* eps */
		int *,		/* kprint */
		int *,		/* inf */
		int *,		/* liw */
		int *,		/* iw */
		int *,		/* lrw */
		double *,	/* rw */
		int *,		/* maxit */
		int *,		/* iscale */
		int *,		/* ichoose */
		int *,		/* iiexact */
		int *,		/* icorr */
		int *,		/* iterations */
		int *);		/* debug option */



#define MINOS_DEBUG FALSE
/* if MINOS_DEBUG then some sanity checking is done in calls from f77 to C */
#define D_ZERO (double)0.0
#define D_ONE (double)1.0

struct count_data {
  int32 used;       /* Number of solver vars/included rels */
};

struct jacobian_data {
  linsol_system_t sys;   /* Jacobian linear system */
  mtx_matrix_t mtx;      /* Matrix from funcon jacobian system */
  mtx_matrix_t objmtx;   /* Matrix from funobj */
  real64 *rhs;     /* RHS from jacobian system */
};
/* NULL ==> not created.  sys, mtxs, and rhs created and destroyed together */

struct slv2_system_structure {
  int integrity;
  slv_parameters_t p;    /* the standard slv parameters and status */
  int iarray[slv2_IA_SIZE];
  double rarray[slv2_RA_SIZE];
  slv_status_t s;
  int panic;   /* bailout called noninteractively in FUNCON/OBJ */
  double clock;  /* cumulative cpu */

/* Problem definition */
   slv_system_t           slv;          /* slv_system_t back-link */
  expr_t obj;                        /* Objective function: NULL = none */
  struct var_variable **vlist;             /* Variable list (NULL terminated) */
  struct var_variable **vlist_user;        /* User vlist (NULL = determine) */
  bnd_boundary_t *blist;             /* Boundary list (NULL terminated) */
  bnd_boundary_t *blist_user;        /* User blist (NULL = none) */
  struct rel_relation **rlist;             /* Relation list (NULL terminated) */
  struct rel_relation **rlist_user;        /* User rlist (NULL = none) */
  struct ExtRelCache **erlist;       /* External rel cache list */
  struct ExtRelCache **erlist_user;  /* User erlist (NULL = none) */

  struct count_data v,r;             /* Variable and relation data */
  int32 maxndx;                /* Maximum index (capacity) */
  int32 mrows;                 /* minos idea of r.used. note that
                                         mrows =1 even if r.used =0 */

/* Arrays allocated at presolve */
  struct jacobian_data jacobian;
  mtx_coord_t *nzlist;   /* Complete list of jacobian elements, C indexed */
  int32
    *iw;		/* integer workspace */
  real64
    *rw, 		/* real workspace */
    *xn, 		/* variable values*/
    *bnds,		/* bound limits */
    *a,  		/* sparse jacobian */
    f,
    *g,
    *c;
  int32
    *avar,
    *acon,
    *ka,		/* column indices for a */
    *hs;		/* variable status (type) vecotr */
  int32
    md,
    nzd,
    ibnds,
    nfunc,
    ngrad,
    iwsize,
    rwsize;

/* scalars */
   int32 itcount;   /* Iteration countdown */
   boolean basis;  /*  basis exists */

/*
 * minos parms
 * fortran files normally 5=stdin, 6=stdout, 0=no file,
 * 9=`pwd`/minos.summary
 */
  int32 ispecs,  	/* fortran input unit number */
               iprint,  /* fortran output unit number */
               isumm,   /* fortran output unit number */
               inform,  /* return flag from opt */
               nb,      /* # free vars + # included rels */
               njac,    /* # jacobian elements (linear+nonlinear) */
               nnobj,  	/* number of nonlinear vars in objective */
	       iter;	/* number of iterations */
};

static slv2_system_t g_sys;   /* Current system: used by FUNOBJ & FUNCON */


/**
 ***  Integrity checks
 ***  ----------------
 ***     check_system(sys)
 **/

#define OK        ((int)492031596)
#define DESTROYED ((int)729104829)
/* note deaddead would be int 3735936685 */
static int check_system(slv2_system_t sys)
/* Checks sys for NULL and for integrity. */
{

  if( sys == NULL ) {
     FPRINTF(stderr,"ERROR:  (slv2) check_system\n");
     FPRINTF(stderr,"        NULL system handle.\n");
     return 1;
  }

  switch( sys->integrity ) {
    case OK:
      return 0;
    case DESTROYED:
      FPRINTF(stderr,"ERROR:  (slv2) check_system\n");
      FPRINTF(stderr,"        System was recently destroyed.\n");
      return 1;
    default:
      FPRINTF(stderr,"ERROR:  (slv2) check_system\n");
      FPRINTF(stderr,"        System reused or never allocated.\n");
      return 1;
  }
}

/*********************************************************************
                  *** Memory management routines ***
free_unless_null(ptr)
zero(arr,nelts,type)
alloc_vector(len)
copy_vector(from,too,len)
free_vector(vec)
make_jacobian(sys)
destroy_jacobian(sys)
*********************************************************************/

static void free_unless_null(POINTER ptr)
{
  if( ptr != NULL )
    ascfree(ptr);
}


#define zero(arr,nelts,type)    \
   mem_zero_byte_cast((arr),0,(nelts)*sizeof(type))
/* Zeros an array of nelts objects, each having given type. */

#define alloc_vector(len) ((real64 *)ascmalloc((len)*sizeof(real64)))
#define alloc_zero_vector(len) ((real64 *)asccalloc((len),sizeof(real64)))
#define alloc_ivector(len) ((int32 *)ascmalloc((len)*sizeof(int32)))
#define alloc_zero_ivector(len) ((int32 *)asccalloc((len),sizeof(int32)))
#define copy_vector(from,too,len)  \
  mem_move_cast((from),(too),(len)*sizeof(real64))
#define free_vector(vec)  \
  free_unless_null((POINTER)(vec))

static void destroy_jacobian(slv2_system_t sys)
/* destroys, deallocates jacobian (sys,mtx, all rhs vectors attached to sys)*/
{
  check_system(sys);
  if( sys->jacobian.sys ) {
    int count = linsol_number_of_rhs(sys->jacobian.sys)-1;
    for( ; count >= 0; count-- )
      free_vector(linsol_get_rhs(sys->jacobian.sys,count));
    mtx_destroy(linsol_get_matrix(sys->jacobian.sys));
    mtx_destroy(sys->jacobian.objmtx);
    linsol_destroy(sys->jacobian.sys);
    sys->jacobian.sys=NULL;
    sys->jacobian.mtx=NULL;
    sys->jacobian.objmtx=NULL;
    sys->jacobian.rhs=NULL;
  }
}

static void make_jacobian(slv2_system_t sys)
/**
 ***  fills linsol, mtx, and one rhs for size sys->maxndx,
 ***  which is assumed set.
 **/
{
  destroy_jacobian(sys);
  sys->jacobian.sys = linsol_create();
  sys->jacobian.mtx = mtx_create();
  sys->jacobian.objmtx = mtx_create();
  mtx_set_order (sys->jacobian.mtx, sys->maxndx );
  mtx_set_order (sys->jacobian.objmtx, sys->maxndx );
  sys->jacobian.rhs = alloc_vector(sys->maxndx);
  linsol_set_matrix(sys->jacobian.sys,sys->jacobian.mtx);
  linsol_add_rhs(sys->jacobian.sys,sys->jacobian.rhs,0);
}
/*********************************************************************
                     *** General I/O routines ***
print_var_name(out,sys,var)
print_bnd_name(out,sys,bnd)
print_rel_name(out,sys,rel)
*********************************************************************/

static void print_vector (slv2_system_t sys, double * vec,int len, char * name)
{
  int i;
  for (i=0; i<len; i++)
  FPRINTF(LIF(sys),"%s(%d) == %20.15g\n",name,i,vec[i]);
  FFLUSH(LIF(sys));
}

static void print_ivector (slv2_system_t sys, int * vec,int len, char * name) {
  int i;
  for (i=0; i<len; i++)
  FPRINTF(LIF(sys),"%s(%d) == %d\n",name,i,vec[i]);
  FFLUSH(LIF(sys));
}

#define print_var_name(a,b,c) slv_print_var_name((a),(b)->slv,(c))
#define print_rel_name(a,b,c) slv_print_rel_name((a),(b)->slv,(c))
#define print_bnd_name(a,b,c) slv_print_bnd_name((a),(c))

/*********************************************************************
iteration_begins(sys)
iteration_ends(sys)
*********************************************************************/

static void iteration_begins(slv2_system_t sys)
/* Prepares sys for an iteration, increasing the iteration counts
   and starting the clock. */
{
  sys->clock = tm_cpu_time();
}

static void iteration_ends(slv2_system_t sys)
/* Prepares sys for exiting an iteration, stopping the clock and recording
   the cpu time, as well as updating the status. */
{
  double cpu_elapsed;   /* elapsed this iteration */
  boolean unsuccessful;

  cpu_elapsed = (double)(tm_cpu_time() - sys->clock);
  sys->s.cpu_elapsed += cpu_elapsed;

  if( !sys->s.converged ) {
    sys->s.block.cpu_elapsed += cpu_elapsed;
    sys->s.time_limit_exceeded =
      (sys->s.block.cpu_elapsed > sys->p.time_limit);
  }

  unsuccessful = sys->s.diverged ||
                 sys->s.inconsistent ||
                 sys->s.iteration_limit_exceeded ||
                 sys->panic ||
                 sys->s.time_limit_exceeded;
/* maybe add this line in a little later */
		 (sys->v.used <= sys->r.used);
  sys->s.ready_to_solve =  !unsuccessful && !sys->s.converged;
  sys->s.ok = !unsuccessful && sys->s.calc_ok && !sys->s.struct_singular;

}

static void install_allvars(slv2_system_t sys,real64 *values)
/*********************************************************************
Moves variables from given array to the value field of each free variable.
*********************************************************************/
{
  int32 col;
  struct var_variable *var;
  for( col=0 ; col < sys->v.used ; ++col ) {
    var = sys->vlist[mtx_col_to_org(sys->jacobian.mtx,col)];
    var_set_value(var,values[col]);
  }
}

static void calc_residuals(slv2_system_t sys)
/* Residuals are calculated, norm value is stored in status, and
   sys->s.calc_ok is set accordingly. Knows nothing about scaling */
{
  real64 sum,res;
  struct rel_relation **rp;

  calc_ok = TRUE;
  sum = D_ZERO;
  for( rp=sys->rlist ; *rp != NULL ; ++rp ) {
    if (rel_included(*rp) && rel_active(*rp)) {
      res = relman_eval(*rp,&calc_ok);
      if( !relman_calc_satisfied(*rp,sys->p.tolerance.feasible) ||
           ( !rel_less(*rp) && !rel_greater(*rp) )
        ) {
        sum += calc_sqr_D0(res);
      }
    }
  }
  if (sum > D_ZERO)
    sys->s.block.residual = calc_sqrt_D0(sum);
  else sys->s.block.residual=sum;

  if( sys->obj != NULL )
    exprman_eval(sys->obj);   /* Check for OK calculations */
  sys->s.calc_ok = calc_ok;
}

/*********************************************************************
                    *** LB_OPT subroutines ***
funobj(...)
*********************************************************************/

static void FUNOBJ_SUB_NAME(int *n,
                int *m,
                double *x,
                double *f,
                double *c)
/*********************************************************************
Computes the objective function and constraint residuals.  Called by LB_OPT.
IN
   n        # of nonlinear obj variables (which is all nl in sys until smarter)
   m        # of nonlinear constraints
   x        Variable values dimension n

OUT
   f        Value of objective function
   c        Vector of constraint residuals
*********************************************************************/
{
  int32 row, eqn_n;
  struct rel_relation **rp;
  rel_filter_t rfilter;

/* shove x back into ASCEND */
  install_allvars(g_sys,x);

  rfilter.included = rel_true;
  rfilter.equality = rel_true;
  rfilter.in_block = rel_ignore;
  rfilter.in_subregion = rel_ignore;

  print_vector(g_sys,x,*n,"x");
  print_vector(g_sys,f,1,"f-in");
  print_vector(g_sys,c,*m,"c-in");

  calc_ok = TRUE;

  *f = (g_sys->obj==NULL) ? D_ZERO : exprman_eval(g_sys->obj);

  FPRINTF(LIF(g_sys),"FUNOBJ returning obj= %g\n",*f);
  if( !calc_ok )  {
    FPRINTF(MIF(g_sys),"!!FUNOBJ: Warning: calculation error(s).\n");
  }

  for( rp=g_sys->rlist; *rp != NULL; rp++ ) {
    eqn_n = rel_sindex(*rp);
    if( rel_apply_filter(*rp,&rfilter) ) {
      g_sys->jacobian.rhs[eqn_n] = relman_eval(*rp,&calc_ok);
    }
  }

  if( !calc_ok ) {
    FPRINTF(MIF(g_sys),"!!FUNOBJ:  Warning:residual calculation error(s).\n");
  }

  /* stuff function results */
  zero(c,*m,double);
  for( row=0 ; row < *m ; ++row )
    c[row] = g_sys->jacobian.rhs[mtx_row_to_org(g_sys->jacobian.mtx,row)];

  iteration_ends(g_sys);  /* timeout to check interface */
  iteration_begins(g_sys);

  print_vector(g_sys,x,*n,"x");
  print_vector(g_sys,f,1,"f-out");
  print_vector(g_sys,c,*m,"c-out");

}

static void FUNGRAD_SUB_NAME(int *m,
                 int *n,
                 int *njac,
		 int *avar,
		 int *acon,
		 int *ka,
                 double *x,
                 double *g,
                 double *a)
/*********************************************************************
Computes the obj. function gradient and constraints jacobian.  Called by LB_OPT.

We really should be scaling the nonlinear residuals. this is lunacy.

For now, don't do energy balances. :-(

Since all variables in non-linear relations/obj are assumed to appear
non-linearly, this routine does not require updating the lvars to be
correct. It will need to update lvars somehow if we sort variables
within a nonlinear equation.

IN
   m        # of nonlinear constraints (rows)
   n        # of nonlinear variables (cols)
   njac     # of varying jacobian elements as known by opt
   avar	    vector of variable indices corresponding to a
   acon	    vector of constraint indices corresponding to a
   ka
   x        nonlinear variable values (index by column)

OUT
   g        Vector of gradients dimension n of objective
   a        Function gradients
*********************************************************************/
{
  int ii; /* for debugging */

  int32 row,ndx,gndx,eqn_n;

  mtx_coord_t nz, *entry;
  real64 value;
  struct rel_relation **rp;
  rel_filter_t rfilter;
  var_filter_t vfilter;
  struct var_variable *var;


  print_vector(g_sys,x,*n,"x");
  print_vector(g_sys,g,*n,"g-in");
  print_vector(g_sys,a,*njac,"a-in");

  install_allvars(g_sys,x);

/* Calculate objective gradients */

  zero(g,*n,double);

  /* calc free & incident dF/dx  */
  vfilter.fixed = var_false;
  vfilter.incident = var_true;
  vfilter.in_block = var_ignore;

  if ( g_sys->obj != NULL ) {

    nz.row=(int32)0;
    mtx_mult_row_zero(g_sys->jacobian.objmtx,nz.row,mtx_ALL_COLS);

    exprman_diffs(g_sys->obj, &vfilter, nz.row, g_sys->jacobian.objmtx);

    /* put in opt gradient vector */
    nz.col = mtx_FIRST;
    while( value = mtx_next_in_row(g_sys->jacobian.objmtx,&nz,mtx_ALL_COLS),
        nz.col != mtx_LAST ) {
      var = g_sys->vlist[mtx_col_to_org(g_sys->jacobian.objmtx,nz.col)];

      g[nz.col] = value;
      FPRINTF(LIF(g_sys),"d(obj)/d(");
      print_var_name(LIF(g_sys),g_sys,var);
      FPRINTF(LIF(g_sys),")=%20.16g\n",g[nz.col]);
      FPRINTF(LIF(g_sys),"g subscript stuffed: %d\n",nz.col);
    } /* while */

    if( !calc_ok ) {
      FPRINTF(MIF(g_sys),"!!FUNGRAD: Warning:gradient calulation error(s).\n");
    }
  }

/* Calculate constraints jacobian */

  rfilter.included = rel_true;
  rfilter.equality = rel_true;
  rfilter.in_block = rel_ignore;
  rfilter.in_subregion = rel_ignore;

  calc_ok = TRUE;

  /* don't zero the nl jacobian a as it also may have constants */
  for( rp=g_sys->rlist; *rp != NULL; rp++ ) {
    eqn_n = rel_sindex(*rp);
    if( rel_apply_filter(*rp,&rfilter) ) {
      mtx_mult_row_zero(g_sys->jacobian.mtx,
                        mtx_org_to_row(g_sys->jacobian.mtx,eqn_n),
                        mtx_ALL_COLS);
      g_sys->jacobian.rhs[eqn_n] =
      relman_diffs(*rp,&vfilter,g_sys->jacobian.mtx);
    }
  }

  if( !calc_ok ) {
    FPRINTF(MIF(g_sys),"!!FUNGRAD: Warning:jacobian calculation error(s).\n");
  }

  FPRINTF(LIF(g_sys),"FUNGRAD jacobian calculated\n");

/*
  for ( ii = 0; ii < *njac; ++ii)
    PRINTF("\nacon[%3d] = %3d, avar[%3d] = %3d", ii,acon[ii], ii,avar[ii]);
 */
/* this may need to be sped up a little later. */
  for (entry = g_sys->nzlist; entry < (g_sys->nzlist + *njac); ++entry) {
    for (ndx = ka[entry->row]; avar[ndx] != (entry->col + 1); ++ndx);
    a[ndx] = mtx_value(g_sys->jacobian.mtx, entry);
  }

  FPRINTF(LIF(g_sys),"FUNGRAD opt jacobian stuffed \n");

  if( !calc_ok ) {
    FPRINTF(MIF(g_sys),"!!FUNGRAD: Warning:jacobian calculation error(s).\n");
  }

  iteration_ends(g_sys); /* timout to check interface */
  iteration_begins(g_sys);

  print_vector(g_sys,x,*n,"x");
  print_vector(g_sys,g,*n,"g-out");
  print_vector(g_sys,a,*njac,"a-out");
}


/*********************************************************************
                     *** LB_OPT interface ***
insure_bounds(sys,var)
make_bounds(sys)
invoke_opt(sys)
*********************************************************************/


#define THEIR_INFINITY (double)1.0e20

static void insure_bounds(slv2_system_t sys,struct var_variable *var)
/**
 ***  Insures that the variable value is within ascend bounds.
 **/
{
  FILE *mif = MIF(sys);
  real64 val,low,high;

  if( sys->p.ignore_bounds )
     return;

  low = var_lower_bound(var);
  high = var_upper_bound(var);
  val = var_value(var);
  if( low > high ) {
    FPRINTF(mif,"Bounds for variable ");
    print_var_name(mif,sys,var);
    FPRINTF(mif," are inconsistent [%g,%g].\n",low,high);
    FPRINTF(mif,"Bounds will be swapped.\n");
    var_set_upper_bound(var, low);
    var_set_lower_bound(var, high);
    low = var_lower_bound(var);
    high = var_upper_bound(var);
  }

  if( low > val ) {
    FPRINTF(mif,"Variable ");
    print_var_name(mif,sys,var);
    FPRINTF(mif," was found below its lower bound.\n");
    FPRINTF(mif,"It will be moved to its lower bound.\n");
    var_set_value(var,low);
  } else if( val > high ) {
    FPRINTF(mif,"Variable ");
    print_var_name(mif,sys,var);
    FPRINTF(mif," was found above its upper bound.\n");
    FPRINTF(mif,"It will be moved to its upper bound.\n");
    var_set_value(var, high);
  }
}

static real64 get_linslack(slv2_system_t sys,
                                 int32 row,
                                 struct rel_relation *rel)
/* This returns the slack of a relation as if the relation were a linear
   equality. When solving a linear system A.x =b with simplex by rearranging
   the system to A.x +s =0, one can either fix RHS=b and drive s->0 or
   fix s=-b (by double bounding) and drive RHS->0. Since Minos doesn't
   have a _convenient_ way to set RHS, we will use the latter approach.
   Simplex lp doesn't notice a difference.
   s(x0)=LHS(x0)-RHS(x0)-A.x0 = resid -Ax. Scaling of A,x won't matter.
   Does not check sanity of rel,row given. nonlinear rels wil give not
   terribly meaningful answers.
   Assumes sys->jacobian.mtx, sys->nzlist stuffed.
   Ignores any fixed variables that have snuck into the jacobian.
*/
{
  real64 s=D_ZERO;
  int ndx;
  struct var_variable *var;

  check_system(sys);
  s=relman_eval(rel,&calc_ok);
  for( ndx=0 ; ndx < sys->njac ; ndx++ ) {
    if( sys->nzlist[ndx].row ==row && sys->nzlist[ndx].col<sys->v.used ) {
      var = sys->vlist[mtx_col_to_org(sys->jacobian.mtx,
                                      sys->nzlist[ndx].col)];
      s += -var_value(var)* mtx_value(sys->jacobian.mtx ,sys->nzlist+ndx);
    }
  }
  return (s);
}

static make_bounds(slv2_system_t sys)
/* Init bnds, xn, a.
   Note: bounds on slacks must be scaled if we start scaling the
   residuals in FUNCON */
{
  int32 col;
  struct var_variable **vp;
  struct var_variable *var;

  for( vp = sys->vlist ; *vp != NULL ; ++vp )
    insure_bounds(sys,*vp);

/* check variable bounds */
  for( col=0 ; col < sys->v.used ; ++col ) {
    var = sys->vlist[mtx_col_to_org(sys->jacobian.mtx,col)];

    if( (sys->bnds[2*col]=var_lower_bound(var)) <= -THEIR_INFINITY )  {
      sys->bnds[2*col]=-THEIR_INFINITY;
      print_var_name(LIF(sys),sys,var);
      FPRINTF(LIF(sys), " has LB_OPT lower bound (%g) tighter than ASCEND's.\n",
         (-THEIR_INFINITY));
    }
    if( (sys->bnds[2*col+1]=var_upper_bound(var)) >= THEIR_INFINITY ) {
      sys->bnds[2*col+1]=THEIR_INFINITY;
      print_var_name(LIF(sys),sys,var);
      FPRINTF(LIF(sys)," LB_OPT upper bound (%g) tighter than ASCEND's.\n",
         THEIR_INFINITY);
    }
  }
  for(col=0; col<sys->v.used; ++col) {
    var = sys->vlist[mtx_col_to_org(sys->jacobian.mtx,col)];
    sys->xn[col] = var_value(var);
  }

  return (0);
}

static void write_inform(slv2_system_t sys)
/* reports opt exit information */
{
  char * res=NULL; /* do not free this ptr ever */
  FILE * fp=MIF(sys);
  FPRINTF(fp,"\n ***** LB_OPT inform= %d ***** \n    ",sys->inform);
  switch (sys->inform) {
    case -10: res="End of iteration.";
	      break;
    case -1 : res="Ready to start a problem.";
	      break;
    case 0  : res="Starting in LB_OPT";
	      break;
    case 1  : res="Optimal solution found.";
              break;
    case 6  : res="User requested termination?";
	      sys->panic = TRUE;
	      break;
    case 10 : res="Optimal solution to square system found.";
	      break;
    case 42 : res="Need to change the size of iw and/or rw";
              break;
    case 55 : res="There is a problem with the bounds on a variable";
	      sys->s.diverged = TRUE;
              break;
    case 57 : res="Basis isn't chosen correctly, check ICHOOSE";
	      sys->s.diverged = TRUE;
              break;
    case 58 : res="Maximum number of iterations exceeded";
	      sys->s.iteration_limit_exceeded = TRUE;
	      break;
    case 73 : res="Line search failure";
	      sys->s.diverged = TRUE;
              break;
    case 78 : res="singular pivot in correction.f";
	      sys->s.diverged = TRUE;
              break;
    case 101 :
    case 102 :
    case 103 :
    case 104 :
    case 105 :
    case 106 :
    case 107 :
    case 108 :
    case 109 :
    case 110 :
    case 111 :
    case 112 :
    case 113 :
    case 114 :
    case 115 :
    case 116 :
    case 117 :
    case 118 :
    case 119 :
    case 120 : res="Iarray or Rarray sizes too small";
	      sys->s.diverged = TRUE;
	      break;
    case 177 :
    case 178 : res="null pointer returned allocing iw or rw";
	      sys->s.diverged = TRUE;
	      break;
    default : res="Unknown inform returned from LB_OPT";
	      sys->panic = TRUE;
              break;
  }
  FPRINTF(fp,"%s\n",res);
  return;
}

static int invoke_opt(slv2_system_t sys)
/* Sets up the interface to LB_OPT and then invokes it. */
{
  static int  i;


  if (!sys->s.ready_to_solve) return (1); /*fail if not presolved*/

  print_vector(sys,sys->xn,sys->v.used,"xn");
  print_vector(sys,sys->a,sys->njac,"a");
  print_vector(sys,sys->bnds,2*sys->v.used,"bnds");
  print_ivector(sys,sys->avar,sys->njac, "avar");
  print_ivector(sys,sys->acon,sys->njac, "acon");

  iteration_begins(sys);

  do {
    if( sys->inform ==  -1 ||
	sys->inform ==   0 ||
	sys->inform == -10 ||
	sys->inform == -15   )
    {
	FUNGRAD_SUB_NAME(&(sys->r.used),        /* m */
		&(sys->v.used),                 /* n */
		&(sys->njac),                   /* nz */
		(sys->avar),                    /* avar */
		(sys->acon),                    /* acon */
		(sys->ka),                      /* ka */
		(sys->xn),                      /* x */
		(sys->g),                       /* g */
		(sys->a));                      /* a */
	sys->ngrad=sys->ngrad + 1;
    }

    if( sys->inform ==  -1 ||
	sys->inform ==   0 ||
	sys->inform == -20   )
    {
	FUNOBJ_SUB_NAME(&(sys->v.used), /* n */
		&(sys->r.used),                 /* m */
		(sys->xn),                      /* x */
		&(sys->f),                      /* f */
		(sys->c));                      /* c */
	sys->nfunc=sys->nfunc + 1;
    }

    if ( sys->iarray[SP2_IDEBUG] ) {
      PRINTF("\nn=%3d\tm=%3d\tmd=%3d\n", sys->v.used,sys->r.used,sys->md);
      if ( sys->v.used <= 100 ) {
	for(i=0; i<sys->v.used; i++)
	  PRINTF("\ti=%3d\tlo=%g\tx[i]=%g\thi=%g\n",
		i,sys->bnds[2*i],sys->xn[i],sys->bnds[2*i+1]);
	for(i=0; i<sys->v.used; i++)
	  PRINTF("\tg[%2d]=%g", i, sys->g[i]);
      }
      PRINTF("\nf=%f\tnjac=%3d\tnzd=%3d\n", sys->f,sys->njac,sys->nzd);

      PRINTF("\n");
      if ( sys->r.used <= 100 ) {
	for(i=0; i<sys->r.used; i++)
	  PRINTF("\tc[%2d]=%g", i, sys->c[i]);
      }

      PRINTF("\n");
      if ( sys->njac <= 100 ) {
	for(i=0; i<sys->njac; i++)
	  PRINTF("\ta[%3d]=%g", i, sys->a[i]);
      }
      PRINTF("\nibnds=%2d\teps=%.8f\n", sys->ibnds,sys->rarray[SP2_EPS]);
      PRINTF("\nkprint=%2d\tinf=%2d", sys->iarray[SP2_KPRINT],sys->inform);
      PRINTF("\nliw=%d\tlrw=%d", sys->iwsize,sys->rwsize);
      PRINTF("\nmaxit=%2d\tichoose=%2d", sys->p.iteration_limit,sys->iarray[SP2_ICHOOSE]);
      PRINTF("\niiexact=%2d\ticorr=%2d", sys->iarray[SP2_IIEXACT],sys->iarray[SP2_ICORR]);
      PRINTF("\niscale=%2d\n", sys->iarray[SP2_ISCALE]);
      PRINTF("\ndebug=%d\n", sys->iarray[SP2_IDEBUG]);
    }

    LB_OPT(&(sys->v.used),			/* n */
	   &(sys->r.used),			/* m */
	   &(sys->md),				/* md */
	   (sys->xn),				/* x */
	   &(sys->f),				/* f */
	   (sys->g),				/* g */
	   (sys->c),				/* c */
	   (sys->a),				/* a */
	   &(sys->njac),			/* nz */
	   &(sys->nzd),				/* nzd */
	   (sys->avar),				/* avar */
	   (sys->acon),				/* acon */
	   &(sys->ibnds),			/* ibnds */
	   (sys->bnds),				/* bnds */
	   &(sys->rarray[SP2_EPS]),		/* eps */
	   &(sys->iarray[SP2_KPRINT]),		/* kprint */
	   &(sys->inform),			/* inf */
	   &(sys->iwsize),			/* liw */
	   (sys->iw),				/* iw */
	   &(sys->rwsize),			/* lrw */
	   (sys->rw),				/* rw */
	   &(sys->p.iteration_limit),		/* maxit */
	   &(sys->iarray[SP2_ISCALE]),		/* iscale */
	   &(sys->iarray[SP2_ICHOOSE]),		/* ichoose */
	   &(sys->iarray[SP2_IIEXACT]),		/* iiexact */
	   &(sys->iarray[SP2_ICORR]),		/* icorr */
	   &(sys->iter),			/* iteration number */
	   &(sys->iarray[SP2_IDEBUG]));		/* debug option */

    if (sys->inform==42) {
      free_vector(sys->iw);
      free_vector(sys->rw);
      PRINTF("\niwsize = %d, rwsize = %d\n", sys->iwsize, sys->rwsize);
      sys->iw=alloc_zero_ivector(sys->iwsize);
      if ( sys->iw == NULL ) {
	PRINTF("\nsys->iw is null, too much memory required");
	sys->inform == 177;
      }
      sys->rw=alloc_zero_vector(sys->rwsize);
      if ( sys->rw == NULL ) {
	PRINTF("\nsys->rw is null, too much memory required");
	sys->inform == 178;
      }
      sys->inform = -1;
      FPRINTF(LIF(sys),
	"LB_OPT workspace %d bytes.\n",(sys->rwsize*sizeof(real64)));
      FPRINTF(LIF(sys),
	"LB_OPT workspace %d bytes.\n",(sys->iwsize*sizeof(int32)));
      continue;
    }

  } while ( (sys->inform !=   1 &&	/* optimal solution found */
	     sys->inform !=  10 &&	/* opt. sol. to sqare sys. found */
	     sys->inform != -10 &&	/* end of one iteration */
	     sys->inform !=  57 &&	/* ichoose incorrect */
	     sys->inform !=  58 &&	/* max iter. reached */
	     sys->inform !=  73 &&	/* line search failure */
	     sys->inform <= 100    ) || /* workspace sizes too small */
	     sys->inform ==  42);	/* reallocate memory */

  iteration_ends(sys);

  PRINTF("\niterations=%d\tfunc.evals=%d\tgrad.evals=%d",sys->iter,sys->nfunc,sys->ngrad);

  if (sys->inform == 1 || sys->inform == 10) {
    sys->s.converged = 1;
    sys->s.block.previous_total_size = sys->v.used;
  }

  sys->s.iteration = sys->nfunc;
  sys->s.block.iteration = sys->ngrad;

  print_vector(sys,sys->xn,sys->v.used,"x");
  print_vector(sys,sys->a,sys->njac,"a");
  print_vector(sys,sys->bnds,2*sys->v.used,"bnds");

  install_allvars(sys,sys->xn);

  sys->basis=TRUE;

  write_inform(sys);

  iteration_begins(sys);
  calc_residuals(sys);
  iteration_ends(sys);
}

/*********************************************************************
                   *** External routines ***
See slv.h
*********************************************************************/

static struct rel_relation **internal_rlist(struct rel_relation * *rlist)
/* Converts rlist to its non-NULL equivalent. */
{
  static struct rel_relation *empty_list[] = {NULL};
  return( rlist==NULL ? empty_list : rlist );
}

slv2_system_t slv2_create()
{
  slv2_system_t sys;

  sys = (slv2_system_t)ascmalloc( sizeof(struct slv2_system_structure) );
  mem_zero_byte_cast( sys , 0 , sizeof(struct slv2_system_structure) );
  sys->integrity = OK;

  /* Only need to initialize non-zeros */
  sys->p.output.more_important = stdout;
  sys->p.time_limit = (double)30.0;
  sys->p.iteration_limit = 100L;
  sys->p.tolerance.singular = (double)1e-8;
  sys->p.tolerance.feasible = (double)1e-6;
  sys->p.tolerance.stationary = (double)1e-8;
  sys->p.partition = TRUE;
  sys->p.whose = slv2_solver_number;
  sys->p.rho = (double)0.1;
  sys->s.ok = TRUE;
  sys->s.calc_ok = TRUE;
  sys->rlist = internal_rlist(NULL);
  sys->iwsize = 10000;
  sys->rwsize = 20000;
  sys->p.sp.iap = &(sys->iarray[0]);
  sys->p.sp.rap = &(sys->rarray[0]);
  sys->iarray[SP2_OPTION]=1;
  sys->iarray[SP2_ISCALE]=1;
  sys->iarray[SP2_ICHOOSE]=0;
  sys->iarray[SP2_ICORR] = 3;
  sys->iarray[SP2_KPRINT]=1;
  sys->iarray[SP2_IIEXACT]=1;
  sys->iarray[SP2_IDEBUG]=0;
  sys->rarray[SP2_EPS]=(double)1.0e-7;
  sys->rarray[SP2_VV]=(double)0.1;
  return(sys);
}


int slv2_destroy(slv2_system_t sys)
{
  if (check_system(sys)) return 1;
  slv2_set_var_list(sys,NULL);      /* also destroys list if internal */
  slv2_set_obj_function(sys,NULL);  /* also destroys obj if internal */
  slv2_set_bnd_list(sys,NULL);
  slv2_set_rel_list(sys,NULL);
  slv2_set_extrel_list(sys,NULL);
  /* the above lists and obj will be destroyed by destroy_from_instance */

  destroy_jacobian(sys);
  free_unless_null( (POINTER)(sys->nzlist) );
  sys->nzlist=NULL;

   /* free int lists */
  free_unless_null( (POINTER)(sys->ka) );
  free_unless_null( (POINTER)(sys->acon) );
  free_unless_null( (POINTER)(sys->avar) );
  free_unless_null( (POINTER)(sys->iw) );
  sys->ka=NULL;
  sys->acon=NULL;
  sys->avar=NULL;
  sys->iw=NULL;

   /* free double lists */
  free_vector(sys->rw);
  free_vector(sys->a);
  free_vector(sys->bnds);
  free_vector(sys->xn);
  sys->xn=NULL;
  sys->bnds=NULL;
  sys->a=NULL;
  sys->rw=NULL;

  sys->integrity = DESTROYED;
  ascfree( (POINTER)sys );
  return 0;
}

void slv2_set_rel_list(slv2_system_t sys,struct rel_relation **rlist)
{
  static struct rel_relation *empty_list[] = {NULL};

  check_system(sys);
  sys->rlist_user = rlist;
  sys->rlist = (rlist == NULL ? empty_list : rlist);
  sys->s.ready_to_solve = FALSE;
}

struct rel_relation **slv2_get_rel_list(slv2_system_t sys)
{
  check_system(sys);
  return( sys->rlist_user );
}

void slv2_set_extrel_list(slv2_system_t sys,struct ExtRelCache **erlist)
{
  static struct ExtRelCache *empty_list[] = {NULL};
  check_system(sys);
  sys->erlist_user = erlist;
  sys->erlist = (erlist==NULL ? empty_list : erlist);
  sys->s.ready_to_solve = FALSE;
}

struct ExtRelCache **slv2_get_extrel_list(slv2_system_t sys)
{
  check_system(sys);
  return( sys->erlist_user );
}

void slv2_set_bnd_list(slv2_system_t sys,bnd_boundary_t *blist)
{
  static bnd_boundary_t empty_list[] = {NULL};

  check_system(sys);
  sys->blist_user = blist;
  sys->blist = (blist == NULL ? empty_list : blist);
  sys->s.ready_to_solve = FALSE;
}

bnd_boundary_t *slv2_get_bnd_list(slv2_system_t sys)
{
  check_system(sys);
  return( sys->blist_user );
}

void slv2_set_var_list(slv2_system_t sys,struct var_variable **vlist)
{
  static struct var_variable *empty_list[] = {NULL};
  check_system(sys);
  if( sys->vlist_user == NULL )
    if( sys->vlist != NULL && pl_length(sys->vlist) > 0 )
      ascfree( (POINTER)(sys->vlist) );
  sys->vlist_user = vlist;
  sys->vlist = (vlist==NULL ? empty_list : vlist);
  sys->s.ready_to_solve = FALSE;
}

struct var_variable **slv2_get_var_list(slv2_system_t sys)
{
  check_system(sys);
  return( sys->vlist_user );
}

void slv2_set_obj_function(slv2_system_t sys,expr_t obj)
{
  check_system(sys);
  sys->obj = obj;
  sys->s.ready_to_solve = FALSE;
}

expr_t slv2_get_obj_function(slv2_system_t sys)
{
  check_system(sys);
  return(sys->obj);
}

int slv2_count_bnds(slv2_system_t sys, bnd_filter_t *bfilter)
{
  bnd_boundary_t *bp;
  int count = 0;
  check_system(sys);
  for( bp=sys->blist; *bp != NULL; bp++ )
    if( bnd_apply_filter(*bp,bfilter) ) ++count;
  return( count );
}

int slv2_count_rels(slv2_system_t sys, rel_filter_t *rfilter)
{
  struct rel_relation **rp;
  int count = 0;
  check_system(sys);
  for( rp=sys->rlist; *rp != NULL; rp++ )
    if( rel_apply_filter(*rp,rfilter) ) ++count;
  return( count );
}

int slv2_count_vars(slv2_system_t sys, var_filter_t *vfilter)
{
  struct var_variable **vp;
  int count = 0;
  check_system(sys);
  for( vp=sys->vlist; *vp != NULL; vp++ )
    if( var_apply_filter(*vp,vfilter) ) ++count;
  return( count );
}

boolean slv2_eligible_solver(slv2_system_t sys)
{
  return( sys->blist_user == NULL );
}

void slv2_get_parameters(slv2_system_t sys,slv_parameters_t *parameters)
{
  check_system(sys);
  mem_copy_cast( &(sys->p) , parameters , sizeof(slv_parameters_t) );
}

void slv2_set_parameters(slv2_system_t sys,slv_parameters_t *parameters)
{
  check_system(sys);
  if (parameters->whose==slv2_solver_number)
  mem_copy_cast( parameters , &(sys->p) , sizeof(slv_parameters_t) );
}

void slv2_get_status(slv2_system_t sys,slv_status_t *status)
{
  check_system(sys);
  mem_copy_cast( &(sys->s) , status , sizeof(slv_status_t) );
}

static struct var_variable **update_vlist(struct var_variable * *vlist,expr_t expr)
/* Updates vlist, adding variables incident on given expression.  The
   old list is destroyed and the new list is returned. */
{
  struct var_variable **newlist,**elist;
  long nelts;

  elist = expr_incidence_list(expr,NULL);
  if( vlist == NULL )
    return(elist);
  if( (nelts=pl_length(elist)) == 0 ) {
    ascfree( (POINTER)elist );
    return(vlist);
  }
  nelts += pl_length(vlist) + 1;
  newlist = (struct var_variable **)ascmalloc( sizeof(struct var_variable *) * (int)nelts );
  pl_merge_0(newlist,vlist,elist);
  ascfree( (POINTER)vlist );
  ascfree( (POINTER)elist );
  return(newlist);
}

static void determine_vlist(slv2_system_t sys)
/* This function constructs the variable list from the relation list.  It
   is assumed that sys->vlist_user == NULL so that this is necessary. */
{
  struct rel_relation **rp;
  bnd_boundary_t *bp;

  if( pl_length(sys->vlist) > 0 )
     ascfree( (POINTER)sys->vlist );
  sys->vlist = NULL;
  for( bp=sys->blist ; *bp != NULL ; ++bp ) {
     sys->vlist = update_vlist(sys->vlist,bnd_lhs(*bp));
     sys->vlist = update_vlist(sys->vlist,bnd_rhs(*bp));
  }
  for( rp=sys->rlist ; *rp != NULL ; ++rp ) {
     sys->vlist = update_vlist(sys->vlist,rel_lhs(*rp));
     sys->vlist = update_vlist(sys->vlist,rel_rhs(*rp));
  }
  if( sys->obj )
     sys->vlist = update_vlist(sys->vlist,sys->obj);
  if( sys->vlist == NULL )
     slv2_set_var_list(sys,NULL);
}

linsol_system_t slv2_get_linsol_sys(slv2_system_t sys)
{
  check_system(sys);
  return sys->jacobian.sys;
}

slv2_dump_internals(slv2_system_t sys,int level)
{
  int i;
  check_system(sys);
/*  for reference:
   struct jacobian_data jacobian;
   real64
     *z,   minos workspace, includes hessian, etc
     *rc,  reduced costs as known to minos
     *pi,  lagrange multipliers as known to minos for nl constraints
     *bl,  lower bound values (v.used)
     *bu,  upper bound values (v.used)
   int32
     *hs;  variable status (type) vector(nb)
*/
  if (level==5) { /*paramvec dump*/
    for (i=0;i<slv2_IA_SIZE;i++)
      FPRINTF(stderr,"iarray[%d]= %d\n",i,sys->iarray[i]);
    for (i=0;i<slv2_RA_SIZE;i++)
      FPRINTF(stderr,"rarray[%d]= %.16g\n",i,sys->rarray[i]);
  }

  if (level==4)  /*nzlist dump*/
    for (i=0;i<sys->njac;i++)
      FPRINTF(stderr,"nzl[%d] r%d c%d\n",
	i,sys->nzlist[i].row,sys->nzlist[i].col);

  if (level==2) { /* dump jacobian */
    FPRINTF(stderr,"jacobian:\n");
    for (i=0;  i<sys->njac; i++)
      FPRINTF(stderr,"nzlist(%d)[%d(ka=%d),col=%d] = %g\n",
	i, sys->nzlist[i].row,sys->ka[i],sys->nzlist[i].col, sys->a[i]);
  }
  if (level==3) { /* dump x, col indexes */
    FPRINTF(stderr,"xn vector:\n");
    for (i=0; i<sys->v.used; i++)
      FPRINTF(stderr,"jcol(%d)-ka>%d  %g < %g < %g\n",
	i,(i<sys->v.used)?sys->ka[i]:-1, sys->bnds[2*i],sys->xn[i],sys->bnds[2*i+1]);
  }
  if (level==1) {
    FPRINTF(stderr,"file channels:\n");
    FPRINTF(stderr,"ispecs = %d\n",sys->ispecs);
    FPRINTF(stderr,"iprint = %d\n",sys->iprint);
    FPRINTF(stderr,"isumm = %d\n",sys->isumm);
    write_inform(sys);
    FPRINTF(stderr," maxndx= %d\n",sys->maxndx);
    FPRINTF(stderr,"r.used (minos m) = %d (%d)\n",sys->r.used,sys->mrows);
    FPRINTF(stderr,"v.used (minoss n) = %d\n",sys->v.used);
    FPRINTF(stderr,"nb= var+con = %d\n",sys->nb);
    FPRINTF(stderr,"njac (opt nz) = %d\n",sys->njac);
    FPRINTF(stderr,"nnobj = %d\n",sys->nnobj);
  }
  if (level>10)
    for (i=1;i<10;i++) slv2_dump_internals(sys,i);
  return (0);
}

/* sort a range (all with the same column index) in nzlist
so that it is increasing row order. predicated on sparsity.
*/

static int nzsort(mtx_coord_t *l, int32 begin, int32 end)
{
  int32 t,curi,curj;
  if (end<begin) {
#if OPT_DEBUG
    FPRINTF(stderr, "Garbage sent to nzsort: begin %d, end %d.",begin,end);
#endif
    return 1;
   /* this implies that a var appears in the objective which does not
      appear in the constraints. */
  }
  t=end-begin;
  if (!t) return 0; /* 1 element -> do nothing */
  if (t==1) { /* if pair, be quick */
    if (l[begin].row > l[end].row ) { /* wrong ordered pair */
      t=l[end].row;
      l[end].row=l[begin].row;
      l[begin].row=t;
    }
    return 0;
  }
  for (curi=begin; curi<end; curi++) /* yes, it's the evil bubble sort */
    for (curj=curi+1; curj<=end; curj++)
      if ( l[curi].row > l[curj].row ) {
        t=l[curj].row;
        l[curj].row=l[curi].row;
        l[curi].row=t;
      }
  return 0;
}
/****
* Presolve creates a sparse jacobian matrix with rows/columns reordered:
* (vars in nonlinear rels)(obj vars not in rels if o nl)(linear vars)(unincid)
* and rows in order (nonlinear rels) (linear rels) (unincluded rels)
*
* The jacobian structure is recorded and manipulated via nzlist, an array
* of matrix coordinates. The coordinates refer to positions in the reordered
* matrix, not the original variable list positions. use row/col_to_org to find
* out what positions in the matrix go with what equations/variables.
* The nzlist structure is used to stuff minos a, the total jacobian, and
* minos g the nonlinear constraint derivatives in FUNCON. The array a
* is indexed by arrays ha, ka which must be correctly set for minos sparse
* jacobian option to work right. Note that the indexes in ha, ka count from
* 1 instead of 0.
*
* Elements of ha must be in ascending minos row order within a column.
* hence, elements of nzlist must also be in increasing mtx column,row
* order.
*
* notes: V.used is the number of vars that would pass filter_var_solver, and
*        hence the coord of the first fixed var, if there are any.
*        V.nonlinear is the number of vars in nonlinear rels/objective, and
*        hence the coord of the first linear var, if there are any.
*
****/
int slv2_presolve(slv2_system_t sys)
{
  struct var_variable **vp;
  var_filter_t vfilter;
  bnd_boundary_t *bp;
  struct rel_relation **rp;
  rel_filter_t rfilter;
  mtx_range_t nl_rows;
  mtx_coord_t nz;
  real64 value;
  int32 col,row,el,ello,elhi,cap, vndx;

  check_system(sys);
  if( sys->vlist == NULL ) {
     FPRINTF(stderr,"ERROR:  (slv2) slv2_presolve\n");
     FPRINTF(stderr,"        Variable list was never set.\n");
     return 1;
  }
  if( sys->blist == NULL ) {
     FPRINTF(stderr,"ERROR:  (slv2) slv2_presolve\n");
     FPRINTF(stderr,"        Boundary list was never set.\n");
     return 1;
  }
  if( sys->rlist == NULL ) {
     FPRINTF(stderr,"ERROR:  (slv2) slv2_presolve\n");
     FPRINTF(stderr,"        Relation list was never set.\n");
     return 1;
  }

/* iteration_begins() does no damage and is needed so */
/* iteration_ends() can be called afterwards.         */
  iteration_begins(sys);

/* Reset global iteration count and cpu-elapsed. */
  sys->s.iteration = 0L;
  sys->s.cpu_elapsed = D_ZERO;
  sys->panic=FALSE;

/* Determine system's extent */
  if( sys->vlist_user == NULL ) /* build one from rel, list */
    determine_vlist(sys);

  sys->maxndx = 0;
  for( vp=sys->vlist,cap=0 ; *vp != NULL ; ++vp ) {
     var_set_sindex(*vp,cap++);
     var_set_in_block(*vp,FALSE);
  }
  sys->maxndx = cap;
  for( rp=sys->rlist,cap=0 ; *rp != NULL ; ++rp ) {
     rel_set_index(*rp,cap++);
     rel_set_in_block(*rp,FALSE);
     rel_set_satisfied(*rp,FALSE);
  }
  sys->maxndx = MAX(sys->maxndx,cap);
  for( bp = sys->blist ; *bp != NULL ; ++bp ) {
     bnd_set_in_block(*bp,FALSE);
     bnd_set_active(*bp,FALSE);
  }

/* Set up jacobian system. sys,mtx,rhs should be created/destroyed together */
  make_jacobian(sys);

  if( sys->obj != NULL )
    exprman_decide_incidence_obj(sys->obj);
  rfilter.included = rel_true;
  rfilter.equality = rel_true;
  rfilter.in_block = rel_ignore;
  rfilter.in_subregion = rel_ignore;
  for( rp=sys->rlist; *rp != NULL; ++rp )
    if( rel_apply_filter(*rp,&rfilter) )
      relman_decide_incidence(*rp);

  vfilter.fixed = var_false;
  vfilter.incident = var_true;
  vfilter.in_block = var_ignore;
  mtx_clear(sys->jacobian.mtx);
  mtx_clear(sys->jacobian.objmtx);
  sys->r.used = 0;
  for( rp=sys->rlist; *rp != NULL; ++rp ) { /* if constraints, nl first */
    rel_set_in_block(*rp,FALSE);
    if( rel_apply_filter(*rp,&rfilter) ) {
      rel_set_in_block(*rp,TRUE);
      relman_get_incidence(*rp,&vfilter,sys->jacobian.mtx);
      mtx_swap_rows(sys->jacobian.mtx,rel_sindex(*rp),sys->r.used);
      ++(sys->r.used);
    }
  }

/* put all the nonlinear rows at the top */
  sys->mrows=(sys->r.used==0)? 1:sys->r.used;

  sys->md = MAX(1, sys->mrows);
  sys->ibnds = 1;

/* This bit assumes all free vars in nonlinear relations occur nonlinearly.  */
/* We need badly a var_linear_in_rel/expression boolean */
/* move nl_vars left in constraints and obj mtx */
  sys->v.used = 0;
  for( vp=sys->vlist ; *vp != NULL ; vp++ ) {
    var_set_in_block(*vp,FALSE);
    if( var_apply_filter(*vp,&vfilter) ) {
      var_set_in_block(*vp,TRUE);
      mtx_swap_cols(sys->jacobian.mtx,var_sindex(*vp),sys->v.used);
      mtx_swap_cols(sys->jacobian.objmtx,var_sindex(*vp),sys->v.used);
      ++(sys->v.used);
    }
  }
    /* go on to the next variable in if */
/* end of for */

/* this bit assumes all free vars in nonlinear objective occur nonlinearly. */
  sys->nnobj=0; /* no nonlinear variables in obj if obj linear or null*/
  if( sys->obj != NULL )
    sys->nnobj=sys->v.used;

/*************************************************************************
   From here on in the MINOS invocation algorithm, don't call clear
   on mtx_matrix_t's or you will screw up the nextinrow/col operators
   used for stuffing gradients. mtx_mult_row if 0s needed.
 *************************************************************************/

 /* count nonzeros */
  sys->njac = 0;
  for( nz.row=0 ; nz.row < sys->maxndx ; ++nz.row )
     sys->njac += mtx_nonzeros_in_row(sys->jacobian.mtx,nz.row,mtx_ALL_COLS);
  /* allocate nzlist with at least one element*/
  free_unless_null( (POINTER)(sys->nzlist) );
  sys->nzlist = (mtx_coord_t *)ascmalloc((sys->njac+1) * sizeof(mtx_coord_t));
  sys->nzlist[sys->njac].row=sys->nzlist[sys->njac].col=sys->r.used+5000000;


  /* copy coord to nzlist for use in calc routines */
  for( sys->njac=nz.row=0 ; nz.row < sys->maxndx ; ++nz.row ) {
    nz.col = mtx_FIRST;
    while( value = mtx_next_in_row(sys->jacobian.mtx,&nz,mtx_ALL_COLS),
	  nz.col != mtx_LAST )
       if( value != D_ZERO )
	  mem_copy_cast(&nz , &(sys->nzlist[sys->njac++]) ,
			sizeof(mtx_coord_t) );
  }


  /* ascmalloc ka, avar, acon */
  free_vector(sys->ka); /* create col offset array  */
  sys->ka=(int *)asccalloc((sys->v.used+1),sizeof(int));/*this is not padding*/

  free_vector(sys->avar); /* create avar array */
  sys->avar=(int *)asccalloc((sys->njac+1),sizeof(int));

  free_vector(sys->acon); /* create acon array */
  sys->acon=(int *)asccalloc((sys->njac+1),sizeof(int));

  free_unless_null( (POINTER)(sys->g) );
  sys->g=(double *)asccalloc((sys->v.used),sizeof(double));

  free_unless_null( (POINTER)(sys->c) );
  sys->c=(double *)asccalloc((sys->r.used),sizeof(double));


  /* translate nzlist into acon, avar, ka */
  /* arys number from 1 in fortran */
  for (nz.col=0; nz.col<sys->njac;nz.col++){
    sys->acon[nz.col]=(sys->nzlist[nz.col].row) + 1;
    sys->avar[nz.col]=(sys->nzlist[nz.col].col) + 1;
  }

  sys->nzd = MAX(1, sys->njac);

  for (nz.col=nz.row=0; nz.col < sys->njac; nz.col++)
    if (sys->nzlist[nz.col].row > nz.row) sys->ka[++nz.row]=nz.col;
  sys->ka[sys->v.used]=sys->njac+1;

  sys->basis =
  sys->s.over_defined =
  sys->s.under_defined =
  sys->s.struct_singular =
  sys->s.converged =
  sys->s.diverged =
  sys->s.inconsistent = FALSE;
  sys->s.block.number_of = 1;
  sys->nb=sys->mrows+sys->v.used;

  sys->s.block.previous_total_size = 0;
  sys->s.block.current_block = 0;
  sys->s.block.current_size = sys->maxndx;
  sys->s.block.iteration = 0L;
  sys->s.block.cpu_elapsed = D_ZERO;
  sys->s.calc_ok = TRUE;

  free_vector(sys->iw); /* create opt integer workspace */
  sys->iw = alloc_zero_ivector(sys->iwsize);
  if (!sys->iw) {
    FPRINTF(stderr,"Error mallocing opt integer workspace array!");
    return (1);
  }

  free_vector(sys->rw); /* create opt real workspace */
  sys->rw = alloc_zero_vector(sys->rwsize);
  if ( !sys->rw) {
    FPRINTF(stderr,"Error mallocing opt real workspace array!");
    return (1);
  }

  free_vector(sys->bnds); /* create bounds array */
  sys->bnds=alloc_zero_vector(2*sys->v.used);
  if (!sys->bnds) {
    FPRINTF(stderr,"Error mallocing bnds array!");
    return (1);
  }

  free_vector(sys->xn); /* create solution array */
  sys->xn=alloc_zero_vector(sys->v.used);
  if (!sys->xn) {
    FPRINTF(stderr,"Error mallocing variable array!");
    return (1);
  }

  free_vector(sys->a); /* create jacobian element array */
  sys->a=alloc_zero_vector(sys->njac+1);
  if (!sys->a) {
    FPRINTF(stderr,"Error mallocing jacobian array!");
    return (1);
  }

  if (make_bounds(sys)>0) {
    FPRINTF(MIF(sys),"Unable to presolve; bad return from make_bounds.\n");
    sys->s.ready_to_solve=FALSE;
    return (1);
  }

  g_sys = sys;

  sys->inform = -1;

  iteration_ends(sys);
}

boolean slv2_change_basis(slv2_system_t sys,int32 var, mtx_range_t *rng){
  return FALSE;
}
/* doesn't work right now */
void slv2_resolve(slv2_system_t sys)
{
  struct rel_relation **rp;
  struct var_variable **vp;

  check_system(sys);

/* iteration_begins() does no damage and is needed so */
/* iteration_ends() can be called afterwards.         */
   iteration_begins(sys);

/* Reset global iteration count and cpu-elapsed. */
  sys->s.iteration = 0L;
  sys->s.cpu_elapsed = D_ZERO;

  for( rp=sys->rlist ; *rp != NULL ; ++rp )
    if( rel_included(*rp) && rel_active(*rp) )
      rel_set_satisfied(*rp,FALSE);

  mtx_clear_region(sys->jacobian.mtx,mtx_ENTIRE_MATRIX);
  sys->s.over_defined =
    sys->s.under_defined =
    sys->s.struct_singular =
    sys->s.converged =
    sys->s.diverged =
    sys->s.inconsistent = FALSE;
  sys->s.block.number_of = 1;

  sys->s.block.previous_total_size = 0;
  sys->s.block.current_block = 0;
  sys->s.block.current_size = sys->maxndx;
  sys->s.block.iteration = 0L;
  sys->s.block.cpu_elapsed = D_ZERO;
  sys->s.calc_ok = TRUE;

  sys->basis = FALSE;

  iteration_ends(sys);
}

void slv2_iterate(slv2_system_t sys)
{
  check_system(sys);
  if( !sys->s.ready_to_solve ) {
    FPRINTF(stderr,"ERROR:  (slv2) slv2_iterate\n");
    FPRINTF(stderr,"	    Not ready to solve.\n");
    return;
  }

  invoke_opt(sys);

}

void slv2_solve(slv2_system_t sys)
{
  while ( sys->s.ready_to_solve )
    slv2_iterate(sys);

}

int slv2_register(SlvFunctionsT *sft)
{
  if (sft==NULL)  {
    FPRINTF(stderr,"slv2_register called with NULL pointer\n");
    return 1;
  }

  sft->name = "Opt";
  sft->traditional_number = "2"; /* THIS IS A TEMPORARY INTERFACE HACK */
  sft->ccreate = slv2_create;
  sft->cdestroy = slv2_destroy;
  sft->celigible = slv2_eligible_solver;
  sft->getparam = slv2_get_parameters;
  sft->setparam = slv2_set_parameters;
  sft->getstatus = slv2_get_status;
  sft->solve = slv2_solve;
  sft->presolve = slv2_presolve;
  sft->iterate = slv2_iterate;
  sft->resolve = slv2_resolve;
  sft->getlinsol = slv2_get_linsol_sys;
  sft->getlinsys = slv2_get_linsolqr_sys;
  sft->getsysmtx = slv2_get_jacobian;
  sft->dumpinternals = slv2_dump_internals;
  return 0;
}
#endif /* #else clause of DYNAMIC_OPTSQP */
#endif /* #else clause of !STATIC_OPTSQP && !DYNAMIC_OPTSQP */
